try:
    import curses

    curses_available = True
except ImportError:
    curses_available = False

import enum
from typing import Tuple

from .vacuum import Vacuum


class Control(enum.Enum):

    Quit = "q"
    Forward = "w"
    ForwardFast = "W"
    Backward = "s"
    BackwardFast = "S"
    Left = "a"
    LeftFast = "A"
    Right = "d"
    RightFast = "D"


class VacuumTUI:
    def __init__(self, vac: Vacuum):
        if not curses_available:
            raise ImportError("curses library is not available")

        self.vac = vac
        self.rot = 0
        self.rot_delta = 30
        self.rot_min = Vacuum.MANUAL_ROTATION_MIN
        self.rot_max = Vacuum.MANUAL_ROTATION_MAX
        self.vel = 0.0
        self.vel_delta = 0.1
        self.vel_min = Vacuum.MANUAL_VELOCITY_MIN
        self.vel_max = Vacuum.MANUAL_VELOCITY_MAX
        self.dur = 10 * 1000

    def run(self) -> None:
        self.vac.manual_start()
        try:
            curses.wrapper(self.main)
        finally:
            self.vac.manual_stop()

    def main(self, screen) -> None:
        screen.addstr("Use wasd to control the device.\n")
        screen.addstr("Hold shift to enable fast mode.\n")
        screen.addstr("Press q to quit.\n")
        screen.refresh()
        self.loop(screen)

    def loop(self, win) -> None:
        done = False
        while not done:
            key = win.getkey()
            text, done = self.handle_key(key)
            win.clear()
            win.addstr(text)
            win.refresh()

    def handle_key(self, key: str) -> Tuple[str, bool]:
        try:
            ctl = Control(key)
        except ValueError as e:
            return "Ignoring %s: %s.\n" % (key, e), False

        done = self.dispatch_control(ctl)
        return self.info(), done

    def dispatch_control(self, ctl: Control) -> bool:
        if ctl == Control.Quit:
            return True

        if ctl == Control.Forward:
            self.vel = min(self.vel + self.vel_delta, self.vel_max)
        elif ctl == Control.ForwardFast:
            self.vel = 0 if self.vel < 0 else self.vel_max

        elif ctl == Control.Backward:
            self.vel = max(self.vel - self.vel_delta, self.vel_min)
        elif ctl == Control.BackwardFast:
            self.vel = 0 if self.vel > 0 else self.vel_min

        elif ctl == Control.Left:
            self.rot = min(self.rot + self.rot_delta, self.rot_max)
        elif ctl == Control.LeftFast:
            self.rot = 0 if self.rot < 0 else self.rot_max

        elif ctl == Control.Right:
            self.rot = max(self.rot - self.rot_delta, self.rot_min)
        elif ctl == Control.RightFast:
            self.rot = 0 if self.rot > 0 else self.rot_min

        self.vac.manual_control(rotation=self.rot, velocity=self.vel, duration=self.dur)
        return False

    def info(self) -> str:
        return "Rotation=%s\nVelocity=%s\n" % (self.rot, self.vel)



import enum
import logging
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .exceptions import DeviceException
from .miot_device import DeviceStatus, MiotDevice

_LOGGER = logging.getLogger(__name__)
_MAPPING = {
    # Source https://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:heater:0000A01A:zhimi-mc2:1
    # Heater (siid=2)
    "power": {"siid": 2, "piid": 1},
    "target_temperature": {"siid": 2, "piid": 5},
    # Countdown (siid=3)
    "countdown_time": {"siid": 3, "piid": 1},
    # Environment (siid=4)
    "temperature": {"siid": 4, "piid": 7},
    # Physical Control Locked (siid=6)
    "child_lock": {"siid": 5, "piid": 1},
    # Alarm (siid=6)
    "buzzer": {"siid": 6, "piid": 1},
    # Indicator light (siid=7)
    "led_brightness": {"siid": 7, "piid": 3},
}

HEATER_PROPERTIES = {
    "temperature_range": (18, 28),
    "delay_off_range": (0, 12 * 3600),
}


class LedBrightness(enum.Enum):
    On = 0
    Off = 1


class HeaterMiotException(DeviceException):
    pass


class HeaterMiotStatus(DeviceStatus):
    """Container for status reports from the Xiaomi Smart Space Heater S."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """
        Response (MIoT format) of Xiaomi Smart Space Heater S (zhimi.heater.mc2):

        [
          { "did": "power", "siid": 2, "piid": 1, "code": 0, "value": False },
          { "did": "target_temperature", "siid": 2, "piid": 5, "code": 0, "value": 18 },
          { "did": "countdown_time", "siid": 3, "piid": 1, "code": 0, "value": 0 },
          { "did": "temperature", "siid": 4, "piid": 7, "code": 0, "value": 22.6 },
          { "did": "child_lock", "siid": 5, "piid": 1, "code": 0, "value": False },
          { "did": "buzzer", "siid": 6, "piid": 1, "code": 0, "value": False },
          { "did": "led_brightness", "siid": 7, "piid": 3, "code": 0, "value": 0 }
        ]
        """
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.is_on else "off"

    @property
    def is_on(self) -> bool:
        """True if device is currently on."""
        return self.data["power"]

    @property
    def target_temperature(self) -> int:
        """Target temperature."""
        return self.data["target_temperature"]

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in seconds."""
        return self.data["countdown_time"]

    @property
    def temperature(self) -> float:
        """Current temperature."""
        return self.data["temperature"]

    @property
    def child_lock(self) -> bool:
        """True if child lock is on, False otherwise."""
        return self.data["child_lock"] is True

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on, False otherwise."""
        return self.data["buzzer"] is True

    @property
    def led_brightness(self) -> LedBrightness:
        """LED indicator brightness."""
        return LedBrightness(self.data["led_brightness"])


class HeaterMiot(MiotDevice):
    """Main class representing the Xiaomi Smart Space Heater S (zhimi.heater.mc2)."""

    mapping = _MAPPING

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Temperature: {result.temperature} °C\n"
            "Target Temperature: {result.target_temperature} °C\n"
            "LED indicator brightness: {result.led_brightness}\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Power-off time: {result.delay_off_countdown} hours\n",
        )
    )
    def status(self) -> HeaterMiotStatus:
        """Retrieve properties."""

        return HeaterMiotStatus(
            {
                prop["did"]: prop["value"] if prop["code"] == 0 else None
                for prop in self.get_properties_for_mapping()
            }
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.set_property("power", True)

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.set_property("power", False)

    @command(
        click.argument("target_temperature", type=int),
        default_output=format_output(
            "Setting target temperature to '{target_temperature}'"
        ),
    )
    def set_target_temperature(self, target_temperature: int):
        """Set target_temperature ."""
        min_temp, max_temp = HEATER_PROPERTIES["temperature_range"]
        if target_temperature < min_temp or target_temperature > max_temp:
            raise HeaterMiotException(
                "Invalid temperature: %s. Must be between %s and %s."
                % (target_temperature, min_temp, max_temp)
            )
        return self.set_property("target_temperature", target_temperature)

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.set_property("child_lock", lock)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.set_property("buzzer", buzzer)

    @command(
        click.argument("brightness", type=EnumType(LedBrightness)),
        default_output=format_output(
            "Setting LED indicator brightness to {brightness}"
        ),
    )
    def set_led_brightness(self, brightness: LedBrightness):
        """Set led brightness."""
        return self.set_property("led_brightness", brightness.value)

    @command(
        click.argument("seconds", type=int),
        default_output=format_output("Setting delayed turn off to {seconds} seconds"),
    )
    def set_delay_off(self, seconds: int):
        """Set delay off seconds."""
        min_delay, max_delay = HEATER_PROPERTIES["delay_off_range"]
        if seconds < min_delay or seconds > max_delay:
            raise HeaterMiotException(
                "Invalid scheduled turn off: %s. Must be between %s and %s"
                % (seconds, min_delay, max_delay)
            )
        return self.set_property("countdown_time", seconds // 3600)



from unittest import TestCase

import pytest

from miio import AirHumidifierMiot
from miio.airhumidifier_miot import (
    AirHumidifierMiotException,
    LedBrightness,
    OperationMode,
    PressedButton,
)

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "power": True,
    "fault": 0,
    "mode": 0,
    "target_humidity": 60,
    "water_level": 32,
    "dry": True,
    "use_time": 2426773,
    "button_pressed": 1,
    "speed_level": 810,
    "temperature": 21.6,
    "fahrenheit": 70.9,
    "humidity": 62,
    "buzzer": False,
    "led_brightness": 1,
    "child_lock": False,
    "motor_speed": 354,
    "actual_speed": 820,
    "power_time": 4272468,
    "clean_mode": False,
}


class DummyAirHumidifierMiot(DummyMiotDevice, AirHumidifierMiot):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_speed": lambda x: self._set_state("speed_level", x),
            "set_target_humidity": lambda x: self._set_state("target_humidity", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_led_brightness": lambda x: self._set_state("led_brightness", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_dry": lambda x: self._set_state("dry", x),
            "set_clean_mode": lambda x: self._set_state("clean_mode", x),
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def airhumidifier(request):
    request.cls.device = DummyAirHumidifierMiot()


@pytest.mark.usefixtures("airhumidifier")
class TestAirHumidifier(TestCase):
    def test_on(self):
        self.device.off()  # ensure off
        assert self.device.status().is_on is False

        self.device.on()
        assert self.device.status().is_on is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.device.status().is_on is True

        self.device.off()
        assert self.device.status().is_on is False

    def test_status(self):
        status = self.device.status()
        assert status.is_on is _INITIAL_STATE["power"]
        assert status.error == _INITIAL_STATE["fault"]
        assert status.mode == OperationMode(_INITIAL_STATE["mode"])
        assert status.target_humidity == _INITIAL_STATE["target_humidity"]
        assert status.water_level == int(_INITIAL_STATE["water_level"] / 1.25)
        assert status.water_tank_detached == (_INITIAL_STATE["water_level"] == 127)
        assert status.dry == _INITIAL_STATE["dry"]
        assert status.use_time == _INITIAL_STATE["use_time"]
        assert status.button_pressed == PressedButton(_INITIAL_STATE["button_pressed"])
        assert status.motor_speed == _INITIAL_STATE["speed_level"]
        assert status.temperature == _INITIAL_STATE["temperature"]
        assert status.fahrenheit == _INITIAL_STATE["fahrenheit"]
        assert status.humidity == _INITIAL_STATE["humidity"]
        assert status.buzzer == _INITIAL_STATE["buzzer"]
        assert status.led_brightness == LedBrightness(_INITIAL_STATE["led_brightness"])
        assert status.child_lock == _INITIAL_STATE["child_lock"]
        assert status.actual_speed == _INITIAL_STATE["actual_speed"]
        assert status.power_time == _INITIAL_STATE["power_time"]

    def test_set_speed(self):
        def speed_level():
            return self.device.status().motor_speed

        self.device.set_speed(200)
        assert speed_level() == 200
        self.device.set_speed(2000)
        assert speed_level() == 2000

        with pytest.raises(AirHumidifierMiotException):
            self.device.set_speed(199)

        with pytest.raises(AirHumidifierMiotException):
            self.device.set_speed(2001)

    def test_set_target_humidity(self):
        def target_humidity():
            return self.device.status().target_humidity

        self.device.set_target_humidity(30)
        assert target_humidity() == 30
        self.device.set_target_humidity(80)
        assert target_humidity() == 80

        with pytest.raises(AirHumidifierMiotException):
            self.device.set_target_humidity(29)

        with pytest.raises(AirHumidifierMiotException):
            self.device.set_target_humidity(81)

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Auto)
        assert mode() == OperationMode.Auto

        self.device.set_mode(OperationMode.Low)
        assert mode() == OperationMode.Low

        self.device.set_mode(OperationMode.Mid)
        assert mode() == OperationMode.Mid

        self.device.set_mode(OperationMode.High)
        assert mode() == OperationMode.High

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_set_dry(self):
        def dry():
            return self.device.status().dry

        self.device.set_dry(True)
        assert dry() is True

        self.device.set_dry(False)
        assert dry() is False

    def test_set_clean_mode(self):
        def clean_mode():
            return self.device.status().clean_mode

        self.device.set_clean_mode(True)
        assert clean_mode() is True

        self.device.set_clean_mode(False)
        assert clean_mode() is False



from unittest import TestCase

import pytest

from miio import AirHumidifier
from miio.airhumidifier import (
    MODEL_HUMIDIFIER_CA1,
    MODEL_HUMIDIFIER_CB1,
    MODEL_HUMIDIFIER_V1,
    AirHumidifierException,
    AirHumidifierStatus,
    LedBrightness,
    OperationMode,
)
from miio.device import DeviceInfo

from .dummies import DummyDevice


class DummyAirHumidifierV1(DummyDevice, AirHumidifier):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_HUMIDIFIER_V1
        self.dummy_device_info = {
            "fw_ver": "1.2.9_5033",
            "token": "68ffffffffffffffffffffffffffffff",
            "otu_stat": [101, 74, 5343, 0, 5327, 407],
            "mmfree": 228248,
            "netif": {
                "gw": "192.168.0.1",
                "localIp": "192.168.0.25",
                "mask": "255.255.255.0",
            },
            "ott_stat": [0, 0, 0, 0],
            "model": "zhimi.humidifier.v1",
            "cfg_time": 0,
            "life": 575661,
            "ap": {"rssi": -35, "ssid": "ap", "bssid": "FF:FF:FF:FF:FF:FF"},
            "wifi_fw_ver": "SD878x-14.76.36.p84-702.1.0-WM",
            "hw_ver": "MW300",
            "ot": "otu",
            "mac": "78:11:FF:FF:FF:FF",
        }
        self.device_info = None

        self.state = {
            "power": "on",
            "mode": "medium",
            "temp_dec": 294,
            "humidity": 33,
            "buzzer": "off",
            "led_b": 2,
            "child_lock": "on",
            "limit_hum": 40,
            "trans_level": 85,
            "use_time": 941100,
            "button_pressed": "led",
            "hw_version": 0,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_led_b": lambda x: self._set_state("led_b", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_limit_hum": lambda x: self._set_state("limit_hum", x),
            "miIO.info": self._get_device_info,
        }
        super().__init__(args, kwargs)

    def _get_device_info(self, _):
        """Return dummy device info."""
        return self.dummy_device_info


@pytest.fixture(scope="class")
def airhumidifierv1(request):
    request.cls.device = DummyAirHumidifierV1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airhumidifierv1")
class TestAirHumidifierV1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        device_info = DeviceInfo(self.device.dummy_device_info)

        assert repr(self.state()) == repr(
            AirHumidifierStatus(self.device.start_state, device_info)
        )

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temp_dec"] / 10.0
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_b"]
        )
        assert self.state().buzzer == (self.device.start_state["buzzer"] == "on")
        assert self.state().child_lock == (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().target_humidity == self.device.start_state["limit_hum"]
        assert self.state().trans_level == self.device.start_state["trans_level"]
        assert self.state().motor_speed is None
        assert self.state().depth is None
        assert self.state().dry is None
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().hardware_version == self.device.start_state["hw_version"]
        assert self.state().button_pressed == self.device.start_state["button_pressed"]

        assert self.state().firmware_version == device_info.firmware_version
        assert (
            self.state().firmware_version_major
            == device_info.firmware_version.rsplit("_", 1)[0]
        )
        assert self.state().firmware_version_minor == int(
            device_info.firmware_version.rsplit("_", 1)[1]
        )
        assert self.state().strong_mode_enabled is False

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Silent)
        assert mode() == OperationMode.Silent

        self.device.set_mode(OperationMode.Medium)
        assert mode() == OperationMode.Medium

        self.device.set_mode(OperationMode.High)
        assert mode() == OperationMode.High

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_led(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led(True)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led(False)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_status_without_temperature(self):
        self.device._reset_state()
        self.device.state["temp_dec"] = None

        assert self.state().temperature is None

    def test_status_without_led_brightness(self):
        self.device._reset_state()
        self.device.state["led_b"] = None

        assert self.state().led_brightness is None

    def test_set_target_humidity(self):
        def target_humidity():
            return self.device.status().target_humidity

        self.device.set_target_humidity(30)
        assert target_humidity() == 30
        self.device.set_target_humidity(60)
        assert target_humidity() == 60
        self.device.set_target_humidity(80)
        assert target_humidity() == 80

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(-1)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(20)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(90)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(110)

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False


class DummyAirHumidifierCA1(DummyDevice, AirHumidifier):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_HUMIDIFIER_CA1
        self.dummy_device_info = {
            "fw_ver": "1.6.6",
            "token": "68ffffffffffffffffffffffffffffff",
            "otu_stat": [101, 74, 5343, 0, 5327, 407],
            "mmfree": 228248,
            "netif": {
                "gw": "192.168.0.1",
                "localIp": "192.168.0.25",
                "mask": "255.255.255.0",
            },
            "ott_stat": [0, 0, 0, 0],
            "model": "zhimi.humidifier.v1",
            "cfg_time": 0,
            "life": 575661,
            "ap": {"rssi": -35, "ssid": "ap", "bssid": "FF:FF:FF:FF:FF:FF"},
            "wifi_fw_ver": "SD878x-14.76.36.p84-702.1.0-WM",
            "hw_ver": "MW300",
            "ot": "otu",
            "mac": "78:11:FF:FF:FF:FF",
        }
        self.device_info = None

        self.state = {
            "power": "on",
            "mode": "medium",
            "temp_dec": 294,
            "humidity": 33,
            "buzzer": "off",
            "led_b": 2,
            "child_lock": "on",
            "limit_hum": 40,
            "use_time": 941100,
            "hw_version": 0,
            # Additional attributes of the zhimi.humidifier.ca1
            "speed": 100,
            "depth": 1,
            "dry": "off",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_led_b": lambda x: self._set_state("led_b", [int(x[0])]),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_limit_hum": lambda x: self._set_state("limit_hum", x),
            "set_dry": lambda x: self._set_state("dry", x),
            "miIO.info": self._get_device_info,
        }
        super().__init__(args, kwargs)

    def _get_device_info(self, _):
        """Return dummy device info."""
        return self.dummy_device_info


@pytest.fixture(scope="class")
def airhumidifierca1(request):
    request.cls.device = DummyAirHumidifierCA1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airhumidifierca1")
class TestAirHumidifierCA1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        device_info = DeviceInfo(self.device.dummy_device_info)

        assert repr(self.state()) == repr(
            AirHumidifierStatus(self.device.start_state, device_info)
        )

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temp_dec"] / 10.0
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_b"]
        )
        assert self.state().buzzer == (self.device.start_state["buzzer"] == "on")
        assert self.state().child_lock == (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().target_humidity == self.device.start_state["limit_hum"]
        assert self.state().trans_level is None
        assert self.state().motor_speed == self.device.start_state["speed"]
        assert self.state().depth == self.device.start_state["depth"]
        assert self.state().water_level == int(self.device.start_state["depth"] / 1.25)
        assert self.state().water_tank_detached == (
            self.device.start_state["depth"] == 127
        )
        assert self.state().dry == (self.device.start_state["dry"] == "on")
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().hardware_version == self.device.start_state["hw_version"]
        assert self.state().button_pressed is None

        assert self.state().firmware_version == device_info.firmware_version
        assert (
            self.state().firmware_version_major
            == device_info.firmware_version.rsplit("_", 1)[0]
        )

        try:
            version_minor = int(device_info.firmware_version.rsplit("_", 1)[1])
        except IndexError:
            version_minor = 0

        assert self.state().firmware_version_minor == version_minor
        assert self.state().strong_mode_enabled is False

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Silent)
        assert mode() == OperationMode.Silent

        self.device.set_mode(OperationMode.Medium)
        assert mode() == OperationMode.Medium

        self.device.set_mode(OperationMode.High)
        assert mode() == OperationMode.High

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_led(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led(True)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led(False)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_status_without_temperature(self):
        self.device._reset_state()
        self.device.state["temp_dec"] = None

        assert self.state().temperature is None

    def test_status_without_led_brightness(self):
        self.device._reset_state()
        self.device.state["led_b"] = None

        assert self.state().led_brightness is None

    def test_set_target_humidity(self):
        def target_humidity():
            return self.device.status().target_humidity

        self.device.set_target_humidity(30)
        assert target_humidity() == 30
        self.device.set_target_humidity(60)
        assert target_humidity() == 60
        self.device.set_target_humidity(80)
        assert target_humidity() == 80

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(-1)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(20)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(90)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(110)

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_set_dry(self):
        def dry():
            return self.device.status().dry

        self.device.set_dry(True)
        assert dry() is True

        self.device.set_dry(False)
        assert dry() is False


class DummyAirHumidifierCB1(DummyDevice, AirHumidifier):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_HUMIDIFIER_CB1
        self.dummy_device_info = {
            "fw_ver": "1.2.9_5033",
            "token": "68ffffffffffffffffffffffffffffff",
            "otu_stat": [101, 74, 5343, 0, 5327, 407],
            "mmfree": 228248,
            "netif": {
                "gw": "192.168.0.1",
                "localIp": "192.168.0.25",
                "mask": "255.255.255.0",
            },
            "ott_stat": [0, 0, 0, 0],
            "model": "zhimi.humidifier.v1",
            "cfg_time": 0,
            "life": 575661,
            "ap": {"rssi": -35, "ssid": "ap", "bssid": "FF:FF:FF:FF:FF:FF"},
            "wifi_fw_ver": "SD878x-14.76.36.p84-702.1.0-WM",
            "hw_ver": "MW300",
            "ot": "otu",
            "mac": "78:11:FF:FF:FF:FF",
        }
        self.device_info = None

        self.state = {
            "power": "on",
            "mode": "medium",
            "humidity": 33,
            "buzzer": "off",
            "led_b": 2,
            "child_lock": "on",
            "limit_hum": 40,
            "use_time": 941100,
            "hw_version": 0,
            # Additional attributes of the zhimi.humidifier.cb1
            "temperature": 29.4,
            "speed": 100,
            "depth": 1,
            "dry": "off",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_led_b": lambda x: self._set_state("led_b", [int(x[0])]),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_limit_hum": lambda x: self._set_state("limit_hum", x),
            "set_dry": lambda x: self._set_state("dry", x),
            "miIO.info": self._get_device_info,
        }
        super().__init__(args, kwargs)

    def _get_device_info(self, _):
        """Return dummy device info."""
        return self.dummy_device_info


@pytest.fixture(scope="class")
def airhumidifiercb1(request):
    request.cls.device = DummyAirHumidifierCB1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("airhumidifiercb1")
class TestAirHumidifierCB1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        device_info = DeviceInfo(self.device.dummy_device_info)

        assert repr(self.state()) == repr(
            AirHumidifierStatus(self.device.start_state, device_info)
        )

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_b"]
        )
        assert self.state().buzzer == (self.device.start_state["buzzer"] == "on")
        assert self.state().child_lock == (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().target_humidity == self.device.start_state["limit_hum"]
        assert self.state().trans_level is None
        assert self.state().motor_speed == self.device.start_state["speed"]
        assert self.state().depth == self.device.start_state["depth"]
        assert self.state().dry == (self.device.start_state["dry"] == "on")
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().hardware_version == self.device.start_state["hw_version"]
        assert self.state().button_pressed is None

        assert self.state().firmware_version == device_info.firmware_version
        assert (
            self.state().firmware_version_major
            == device_info.firmware_version.rsplit("_", 1)[0]
        )
        assert self.state().firmware_version_minor == int(
            device_info.firmware_version.rsplit("_", 1)[1]
        )
        assert self.state().strong_mode_enabled is False

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Silent)
        assert mode() == OperationMode.Silent

        self.device.set_mode(OperationMode.Medium)
        assert mode() == OperationMode.Medium

        self.device.set_mode(OperationMode.High)
        assert mode() == OperationMode.High

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_led(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led(True)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led(False)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_status_without_temperature(self):
        self.device._reset_state()
        self.device.state["temperature"] = None

        assert self.state().temperature is None

    def test_status_without_led_brightness(self):
        self.device._reset_state()
        self.device.state["led_b"] = None

        assert self.state().led_brightness is None

    def test_set_target_humidity(self):
        def target_humidity():
            return self.device.status().target_humidity

        self.device.set_target_humidity(30)
        assert target_humidity() == 30
        self.device.set_target_humidity(60)
        assert target_humidity() == 60
        self.device.set_target_humidity(80)
        assert target_humidity() == 80

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(-1)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(20)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(90)

        with pytest.raises(AirHumidifierException):
            self.device.set_target_humidity(110)

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_set_dry(self):
        def dry():
            return self.device.status().dry

        self.device.set_dry(True)
        assert dry() is True

        self.device.set_dry(False)
        assert dry() is False



from unittest import TestCase

import pytest

from miio import Yeelight
from miio.yeelight import YeelightException, YeelightMode, YeelightStatus

from .dummies import DummyDevice


class DummyLight(DummyDevice, Yeelight):
    def __init__(self, *args, **kwargs):
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_bright": lambda x: self._set_state("bright", x),
            "set_ct_abx": lambda x: self._set_state("ct", x),
            "set_rgb": lambda x: self._set_state("rgb", x),
            "set_hsv": lambda x: self._set_state("hsv", x),
            "set_name": lambda x: self._set_state("name", x),
            "set_ps": lambda x: self.set_config(x),
            "toggle": self.toggle_power,
            "set_default": lambda x: "ok",
        }

        super().__init__(*args, **kwargs)

    def set_config(self, x):
        key, value = x
        config_mapping = {"cfg_lan_ctrl": "lan_ctrl", "cfg_save_state": "save_state"}

        self._set_state(config_mapping[key], [value])

    def toggle_power(self, _):
        if self.state["power"] == "on":
            self.state["power"] = "off"
        else:
            self.state["power"] = "on"


class DummyCommonBulb(DummyLight):
    def __init__(self, *args, **kwargs):
        self.state = {
            "name": "test name",
            "lan_ctrl": "1",
            "save_state": "1",
            "delayoff": "0",
            "music_on": "1",
            "power": "off",
            "bright": "100",
            "color_mode": "2",
            "rgb": "",
            "hue": "",
            "sat": "",
            "ct": "3584",
            "flowing": "",
            "flow_params": "",
            "active_mode": "",
            "nl_br": "",
            "bg_power": "",
            "bg_bright": "",
            "bg_lmode": "",
            "bg_rgb": "",
            "bg_hue": "",
            "bg_sat": "",
            "bg_ct": "",
            "bg_flowing": "",
            "bg_flow_params": "",
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="class")
def dummycommonbulb(request):
    request.cls.device = DummyCommonBulb()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("dummycommonbulb")
class TestYeelightCommon(TestCase):
    def test_on(self):
        self.device.off()  # make sure we are off
        assert self.device.status().is_on is False

        self.device.on()
        assert self.device.status().is_on is True

    def test_off(self):
        self.device.on()  # make sure we are on
        assert self.device.status().is_on is True

        self.device.off()
        assert self.device.status().is_on is False

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(50)
        assert brightness() == 50
        self.device.set_brightness(0)
        assert brightness() == 0
        self.device.set_brightness(100)

        with pytest.raises(YeelightException):
            self.device.set_brightness(-100)

        with pytest.raises(YeelightException):
            self.device.set_brightness(200)

    def test_set_color_temp(self):
        def color_temp():
            return self.device.status().color_temp

        self.device.set_color_temp(2000)
        assert color_temp() == 2000
        self.device.set_color_temp(6500)
        assert color_temp() == 6500

        with pytest.raises(YeelightException):
            self.device.set_color_temp(1000)

        with pytest.raises(YeelightException):
            self.device.set_color_temp(7000)

    def test_set_developer_mode(self):
        def dev_mode():
            return self.device.status().developer_mode

        orig_mode = dev_mode()
        self.device.set_developer_mode(not orig_mode)
        new_mode = dev_mode()
        assert new_mode is not orig_mode
        self.device.set_developer_mode(not new_mode)
        assert new_mode is not dev_mode()

    def test_set_save_state_on_change(self):
        def save_state():
            return self.device.status().save_state_on_change

        orig_state = save_state()
        self.device.set_save_state_on_change(not orig_state)
        new_state = save_state()
        assert new_state is not orig_state
        self.device.set_save_state_on_change(not new_state)
        new_state = save_state()
        assert new_state is orig_state

    def test_set_name(self):
        def name():
            return self.device.status().name

        assert name() == "test name"
        self.device.set_name("new test name")
        assert name() == "new test name"

    def test_toggle(self):
        def is_on():
            return self.device.status().is_on

        orig_state = is_on()
        self.device.toggle()
        new_state = is_on()
        assert orig_state != new_state

        self.device.toggle()
        new_state = is_on()
        assert new_state == orig_state

    @pytest.mark.skip("cannot be tested easily")
    def test_set_default(self):
        self.fail()

    @pytest.mark.skip("set_scene is not implemented")
    def test_set_scene(self):
        self.fail()


class DummyLightСolor(DummyLight):
    def __init__(self, *args, **kwargs):
        self.state = {
            "name": "test name",
            "lan_ctrl": "1",
            "save_state": "1",
            "delayoff": "0",
            "music_on": "1",
            "power": "off",
            "bright": "100",
            "color_mode": "2",
            "rgb": "16711680",
            "hue": "359",
            "sat": "100",
            "ct": "3584",
            "flowing": "0",
            "flow_params": "0,0,1000,1,16711680,100,1000,1,65280,100,1000,1,255,100",
            "active_mode": "",
            "nl_br": "",
            "bg_power": "",
            "bg_bright": "",
            "bg_lmode": "",
            "bg_rgb": "",
            "bg_hue": "",
            "bg_sat": "",
            "bg_ct": "",
            "bg_flowing": "",
            "bg_flow_params": "",
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="class")
def dummylightcolor(request):
    request.cls.device = DummyLightСolor()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("dummylightcolor")
class TestYeelightLightColor(TestCase):
    def test_status(self):
        self.device._reset_state()
        status = self.device.status()  # type: YeelightStatus

        assert repr(status) == repr(YeelightStatus(self.device.start_state))

        assert status.name == self.device.start_state["name"]
        assert status.developer_mode is True
        assert status.save_state_on_change is True
        assert status.delay_off == 0
        assert status.music_mode is True
        assert len(status.lights) == 1
        assert status.is_on is False and status.is_on == status.lights[0].is_on
        assert (
            status.brightness == 100
            and status.brightness == status.lights[0].brightness
        )
        assert (
            status.color_mode == YeelightMode.ColorTemperature
            and status.color_mode == status.lights[0].color_mode
        )
        assert (
            status.color_temp == 3584
            and status.color_temp == status.lights[0].color_temp
        )
        assert status.rgb is None and status.rgb == status.lights[0].rgb
        assert status.hsv is None and status.hsv == status.lights[0].hsv
        # following are tested in set mode tests
        # assert status.rgb == 16711680
        # assert status.hsv == (359, 100, 100)
        assert (
            status.color_flowing is False
            and status.color_flowing == status.lights[0].color_flowing
        )
        assert (
            status.color_flow_params is None
            and status.color_flow_params == status.lights[0].color_flow_params
        )
        # color_flow_params will be tested after future implementation
        # assert status.color_flow_params == "0,0,1000,1,16711680,100,1000,1,65280,100,1000,1,255,100" and status.color_flow_params == status.lights[0].color_flow_params
        assert status.moonlight_mode is None
        assert status.moonlight_mode_brightness is None

    def test_set_rgb(self):
        def rgb():
            return self.device.status().rgb

        self.device._reset_state()
        self.device._set_state("color_mode", [1])

        assert rgb() == (255, 0, 0)

        self.device.set_rgb((0, 0, 1))
        assert rgb() == (0, 0, 1)
        self.device.set_rgb((255, 255, 0))
        assert rgb() == (255, 255, 0)
        self.device.set_rgb((255, 255, 255))
        assert rgb() == (255, 255, 255)

        with pytest.raises(YeelightException):
            self.device.set_rgb((-1, 0, 0))

        with pytest.raises(YeelightException):
            self.device.set_rgb((256, 0, 0))

        with pytest.raises(YeelightException):
            self.device.set_rgb((0, -1, 0))

        with pytest.raises(YeelightException):
            self.device.set_rgb((0, 256, 0))

        with pytest.raises(YeelightException):
            self.device.set_rgb((0, 0, -1))

        with pytest.raises(YeelightException):
            self.device.set_rgb((0, 0, 256))

    @pytest.mark.skip("hsv is not properly implemented")
    def test_set_hsv(self):
        self.reset_state()
        hue, sat, val = self.device.status().hsv
        assert hue == 359
        assert sat == 100
        assert val == 100

        self.device.set_hsv()


class DummyLightCeilingV1(DummyLight):  # without background light
    def __init__(self, *args, **kwargs):
        self.state = {
            "name": "test name",
            "lan_ctrl": "1",
            "save_state": "1",
            "delayoff": "0",
            "music_on": "",
            "power": "off",
            "bright": "100",
            "color_mode": "2",
            "rgb": "",
            "hue": "",
            "sat": "",
            "ct": "3584",
            "flowing": "0",
            "flow_params": "0,0,2000,3,0,33,2000,3,0,100",
            "active_mode": "1",
            "nl_br": "100",
            "bg_power": "",
            "bg_bright": "",
            "bg_lmode": "",
            "bg_rgb": "",
            "bg_hue": "",
            "bg_sat": "",
            "bg_ct": "",
            "bg_flowing": "",
            "bg_flow_params": "",
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="class")
def dummylightceilingv1(request):
    request.cls.device = DummyLightCeilingV1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("dummylightceilingv1")
class TestYeelightLightCeilingV1(TestCase):
    def test_status(self):
        self.device._reset_state()
        status = self.device.status()  # type: YeelightStatus

        assert repr(status) == repr(YeelightStatus(self.device.start_state))

        assert status.name == self.device.start_state["name"]
        assert status.developer_mode is True
        assert status.save_state_on_change is True
        assert status.delay_off == 0
        assert status.music_mode is None
        assert len(status.lights) == 1
        assert status.is_on is False and status.is_on == status.lights[0].is_on
        assert (
            status.brightness == 100
            and status.brightness == status.lights[0].brightness
        )
        assert (
            status.color_mode == YeelightMode.ColorTemperature
            and status.color_mode == status.lights[0].color_mode
        )
        assert (
            status.color_temp == 3584
            and status.color_temp == status.lights[0].color_temp
        )
        assert status.rgb is None and status.rgb == status.lights[0].rgb
        assert status.hsv is None and status.hsv == status.lights[0].hsv
        # following are tested in set mode tests
        # assert status.rgb == 16711680
        # assert status.hsv == (359, 100, 100)
        assert (
            status.color_flowing is False
            and status.color_flowing == status.lights[0].color_flowing
        )
        assert (
            status.color_flow_params is None
            and status.color_flow_params == status.lights[0].color_flow_params
        )
        # color_flow_params will be tested after future implementation
        # assert status.color_flow_params == "0,0,1000,1,16711680,100,1000,1,65280,100,1000,1,255,100" and status.color_flow_params == status.lights[0].color_flow_params
        assert status.moonlight_mode is True
        assert status.moonlight_mode_brightness == 100


class DummyLightCeilingV2(DummyLight):  # without background light
    def __init__(self, *args, **kwargs):
        self.state = {
            "name": "test name",
            "lan_ctrl": "1",
            "save_state": "1",
            "delayoff": "0",
            "music_on": "",
            "power": "off",
            "bright": "100",
            "color_mode": "2",
            "rgb": "",
            "hue": "",
            "sat": "",
            "ct": "3584",
            "flowing": "0",
            "flow_params": "0,0,2000,3,0,33,2000,3,0,100",
            "active_mode": "1",
            "nl_br": "100",
            "bg_power": "off",
            "bg_bright": "100",
            "bg_lmode": "2",
            "bg_rgb": "15531811",
            "bg_hue": "65",
            "bg_sat": "86",
            "bg_ct": "4000",
            "bg_flowing": "0",
            "bg_flow_params": "0,0,3000,4,16711680,100,3000,4,65280,100,3000,4,255,100",
        }
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="class")
def dummylightceilingv2(request):
    request.cls.device = DummyLightCeilingV2()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("dummylightceilingv2")
class TestYeelightLightCeilingV2(TestCase):
    def test_status(self):
        self.device._reset_state()
        status = self.device.status()  # type: YeelightStatus

        assert repr(status) == repr(YeelightStatus(self.device.start_state))

        assert status.name == self.device.start_state["name"]
        assert status.developer_mode is True
        assert status.save_state_on_change is True
        assert status.delay_off == 0
        assert status.music_mode is None
        assert len(status.lights) == 2
        assert status.is_on is False and status.is_on == status.lights[0].is_on
        assert (
            status.brightness == 100
            and status.brightness == status.lights[0].brightness
        )
        assert (
            status.color_mode == YeelightMode.ColorTemperature
            and status.color_mode == status.lights[0].color_mode
        )
        assert (
            status.color_temp == 3584
            and status.color_temp == status.lights[0].color_temp
        )
        assert status.rgb is None and status.rgb == status.lights[0].rgb
        assert status.hsv is None and status.hsv == status.lights[0].hsv
        # following are tested in set mode tests
        # assert status.rgb == 16711680
        # assert status.hsv == (359, 100, 100)
        assert (
            status.color_flowing is False
            and status.color_flowing == status.lights[0].color_flowing
        )
        assert (
            status.color_flow_params is None
            and status.color_flow_params == status.lights[0].color_flow_params
        )
        # color_flow_params will be tested after future implementation
        # assert status.color_flow_params == "0,0,1000,1,16711680,100,1000,1,65280,100,1000,1,255,100" and status.color_flow_params == status.lights[0].color_flow_params
        assert status.lights[1].is_on is False
        assert status.lights[1].brightness == 100
        assert status.lights[1].color_mode == YeelightMode.ColorTemperature
        assert status.lights[1].color_temp == 4000
        assert status.lights[1].rgb is None
        assert status.lights[1].hsv is None
        # following are tested in set mode tests
        # assert status.rgb == 15531811
        # assert status.hsv == (65, 86, 100)
        assert status.lights[1].color_flowing is False
        assert status.lights[1].color_flow_params is None
        assert status.moonlight_mode is True
        assert status.moonlight_mode_brightness == 100



"""miIO protocol implementation.

This module contains the implementation of routines to send handshakes, send commands
and discover devices (MiIOProtocol).
"""
import binascii
import codecs
import logging
import socket
from datetime import datetime, timedelta
from typing import Any, Dict, List

import construct

from .exceptions import DeviceError, DeviceException, RecoverableError
from .protocol import Message

_LOGGER = logging.getLogger(__name__)


class MiIOProtocol:
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        timeout: int = 5,
    ) -> None:
        """Create a :class:`Device` instance.

        :param ip: IP address or a hostname for the device
        :param token: Token used for encryption
        :param start_id: Running message id sent to the device
        :param debug: Wanted debug level
        """
        self.ip = ip
        self.port = 54321
        if token is None:
            token = 32 * "0"
        self.token = bytes.fromhex(token)
        self.debug = debug
        self.lazy_discover = lazy_discover
        self._timeout = timeout
        self.__id = start_id

        self._discovered = False
        # these come from the device, but we initialize them here to make mypy happy
        self._device_ts: datetime = datetime.utcnow()
        self._device_id = bytes()

    def send_handshake(self, *, retry_count=3) -> Message:
        """Send a handshake to the device.

        This returns some information, such as device type and serial,
        as well as device's timestamp in response.

        The handshake must also be done regularly to enable communication
        with the device.

        :raises DeviceException: if the device could not be discovered after retries.
        """
        try:
            m = MiIOProtocol.discover(self.ip)
        except DeviceException as ex:
            if retry_count > 0:
                return self.send_handshake(retry_count=retry_count - 1)

            raise ex

        if m is None:
            _LOGGER.debug("Unable to discover a device at address %s", self.ip)
            raise DeviceException("Unable to discover the device %s" % self.ip)

        header = m.header.value
        self._device_id = header.device_id
        self._device_ts = header.ts
        self._discovered = True

        if self.debug > 1:
            _LOGGER.debug(m)
        _LOGGER.debug(
            "Discovered %s with ts: %s, token: %s",
            binascii.hexlify(self._device_id).decode(),
            self._device_ts,
            codecs.encode(m.checksum, "hex"),
        )

        return m

    @staticmethod
    def discover(addr: str = None, timeout: int = 5) -> Any:
        """Scan for devices in the network. This method is used to discover supported
        devices by sending a handshake message to the broadcast address on port 54321.
        If the target IP address is given, the handshake will be send as an unicast
        packet.

        :param str addr: Target IP address
        """
        is_broadcast = addr is None
        seen_addrs = []  # type: List[str]
        if is_broadcast:
            addr = "<broadcast>"
            is_broadcast = True
            _LOGGER.info("Sending discovery to %s with timeout of %ss..", addr, timeout)
        # magic, length 32
        helobytes = bytes.fromhex(
            "21310020ffffffffffffffffffffffffffffffffffffffffffffffffffffffff"
        )

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        s.settimeout(timeout)
        for _ in range(3):
            s.sendto(helobytes, (addr, 54321))
        while True:
            try:
                data, recv_addr = s.recvfrom(1024)
                m = Message.parse(data)  # type: Message
                _LOGGER.debug("Got a response: %s", m)
                if not is_broadcast:
                    return m

                if recv_addr[0] not in seen_addrs:
                    _LOGGER.info(
                        "  IP %s (ID: %s) - token: %s",
                        recv_addr[0],
                        binascii.hexlify(m.header.value.device_id).decode(),
                        codecs.encode(m.checksum, "hex"),
                    )
                    seen_addrs.append(recv_addr[0])
            except socket.timeout:
                if is_broadcast:
                    _LOGGER.info("Discovery done")
                return  # ignore timeouts on discover
            except Exception as ex:
                _LOGGER.warning("error while reading discover results: %s", ex)
                break

    def send(
        self,
        command: str,
        parameters: Any = None,
        retry_count: int = 3,
        *,
        extra_parameters: Dict = None
    ) -> Any:
        """Build and send the given command. Note that this will implicitly call
        :func:`send_handshake` to do a handshake, and will re-try in case of errors
        while incrementing the `_id` by 100.

        :param str command: Command to send
        :param dict parameters: Parameters to send, or an empty list
        :param retry_count: How many times to retry in case of failure, how many handshakes to send
        :param dict extra_parameters: Extra top-level parameters
        :raises DeviceException: if an error has occurred during communication.
        """

        if not self.lazy_discover or not self._discovered:
            self.send_handshake()

        request = self._create_request(command, parameters, extra_parameters)

        send_ts = self._device_ts + timedelta(seconds=1)
        header = {
            "length": 0,
            "unknown": 0x00000000,
            "device_id": self._device_id,
            "ts": send_ts,
        }

        msg = {"data": {"value": request}, "header": {"value": header}, "checksum": 0}
        m = Message.build(msg, token=self.token)
        _LOGGER.debug("%s:%s >>: %s", self.ip, self.port, request)
        if self.debug > 1:
            _LOGGER.debug(
                "send (timeout %s): %s",
                self._timeout,
                Message.parse(m, token=self.token),
            )

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(self._timeout)

        try:
            s.sendto(m, (self.ip, self.port))
        except OSError as ex:
            _LOGGER.error("failed to send msg: %s", ex)
            raise DeviceException from ex

        try:
            data, addr = s.recvfrom(4096)
            m = Message.parse(data, token=self.token)

            header = m.header.value
            payload = m.data.value

            self.__id = payload["id"]
            self._device_ts = header["ts"]  # type: ignore  # ts uses timeadapter

            if self.debug > 1:
                _LOGGER.debug("recv from %s: %s", addr[0], m)

            _LOGGER.debug(
                "%s:%s (ts: %s, id: %s) << %s",
                self.ip,
                self.port,
                header["ts"],
                payload["id"],
                payload,
            )
            if "error" in payload:
                self._handle_error(payload["error"])

            try:
                return payload["result"]
            except KeyError:
                return payload
        except construct.core.ChecksumError as ex:
            raise DeviceException(
                "Got checksum error which indicates use "
                "of an invalid token. "
                "Please check your token!"
            ) from ex
        except OSError as ex:
            if retry_count > 0:
                _LOGGER.debug(
                    "Retrying with incremented id, retries left: %s", retry_count
                )
                self.__id += 100
                self._discovered = False
                return self.send(
                    command,
                    parameters,
                    retry_count - 1,
                    extra_parameters=extra_parameters,
                )

            _LOGGER.error("Got error when receiving: %s", ex)
            raise DeviceException("No response from the device") from ex

        except RecoverableError as ex:
            if retry_count > 0:
                _LOGGER.debug(
                    "Retrying to send failed command, retries left: %s", retry_count
                )
                return self.send(
                    command,
                    parameters,
                    retry_count - 1,
                    extra_parameters=extra_parameters,
                )

            _LOGGER.error("Got error when receiving: %s", ex)
            raise DeviceException("Unable to recover failed command") from ex

    @property
    def _id(self) -> int:
        """Increment and return the sequence id."""
        self.__id += 1
        if self.__id >= 9999:
            self.__id = 1
        return self.__id

    @property
    def raw_id(self):
        return self.__id

    def _handle_error(self, error):
        """Raise exception based on the given error code."""
        if "code" in error and error["code"] == -30001:
            raise RecoverableError(error)
        raise DeviceError(error)

    def _create_request(
        self, command: str, parameters: Any, extra_parameters: Dict = None
    ):
        """Create request payload."""
        request = {"id": self._id, "method": command}

        if parameters is not None:
            request["params"] = parameters
        else:
            request["params"] = []

        if extra_parameters is not None:
            request = {**request, **extra_parameters}

        return request



import logging
from collections import defaultdict
from typing import Any, Dict, List, Tuple

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException
from .utils import int_to_rgb

_LOGGER = logging.getLogger(__name__)


class PhilipsMoonlightException(DeviceException):
    pass


class PhilipsMoonlightStatus(DeviceStatus):
    """Container for status reports from Xiaomi Philips Zhirui Bedside Lamp."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a Moonlight (philips.light.moonlight):

        {'pow': 'off', 'sta': 0, 'bri': 1, 'rgb': 16741971, 'cct': 1, 'snm': 0, 'spr': 0,
         'spt': 15, 'wke': 0, 'bl': 1, 'ms': 1, 'mb': 1, 'wkp': [0, 24, 0]}
        """
        self.data = data

    @property
    def power(self) -> str:
        return self.data["pow"]

    @property
    def is_on(self) -> bool:
        return self.power == "on"

    @property
    def brightness(self) -> int:
        return self.data["bri"]

    @property
    def color_temperature(self) -> int:
        return self.data["cct"]

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """Return color in RGB."""
        return int_to_rgb(int(self.data["rgb"]))

    @property
    def scene(self) -> int:
        return self.data["snm"]

    @property
    def sleep_assistant(self) -> int:
        """Example values:

        0: Unknown
        1: Unknown
        2: Sleep assistant enabled
        3: Awake
        """
        return self.data["sta"]

    @property
    def sleep_off_time(self) -> int:
        return self.data["spr"]

    @property
    def total_assistant_sleep_time(self) -> int:
        return self.data["spt"]

    @property
    def brand_sleep(self) -> bool:
        # sp_sleep_open?
        return self.data["ms"] == 1

    @property
    def brand(self) -> bool:
        # sp_xm_bracelet?
        return self.data["mb"] == 1

    @property
    def wake_up_time(self) -> List[int]:
        # Example: [weekdays?, hour, minute]
        return self.data["wkp"]


class PhilipsMoonlight(Device):
    """Main class representing Xiaomi Philips Zhirui Bedside Lamp.

    Not yet implemented features/methods:

    add_mb                          # Add miband
    get_band_period                 # Bracelet work time
    get_mb_rssi                     # Miband RSSI
    get_mb_mac                      # Miband MAC address
    enable_mibs
    set_band_period
    miIO.bleStartSearchBand
    miIO.bleGetNearbyBandList

    enable_sub_voice                # Sub voice control?
    enable_voice                    # Voice control

    skip_breath
    set_sleep_time
    set_wakeup_time
    en_sleep
    en_wakeup
    go_night                        # Night light / read mode
    get_wakeup_time
    enable_bl                       # Night light
    """

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Brightness: {result.brightness}\n"
            "Color temperature: {result.color_temperature}\n"
            "RGB: {result.rgb}\n"
            "Scene: {result.scene}\n",
        )
    )
    def status(self) -> PhilipsMoonlightStatus:
        """Retrieve properties."""
        properties = [
            "pow",
            "sta",
            "bri",
            "rgb",
            "cct",
            "snm",
            "spr",
            "spt",
            "wke",
            "bl",
            "ms",
            "mb",
            "wkp",
        ]
        values = self.get_properties(properties)

        return PhilipsMoonlightStatus(
            defaultdict(lambda: None, zip(properties, values))
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("rgb", default=[255] * 3, type=click.Tuple([int, int, int])),
        default_output=format_output("Setting color to {rgb}"),
    )
    def set_rgb(self, rgb: Tuple[int, int, int]):
        """Set color in RGB."""
        for color in rgb:
            if color < 0 or color > 255:
                raise PhilipsMoonlightException("Invalid color: %s" % color)

        return self.send("set_rgb", [*rgb])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting brightness to {level}"),
    )
    def set_brightness(self, level: int):
        """Set brightness level."""
        if level < 1 or level > 100:
            raise PhilipsMoonlightException("Invalid brightness: %s" % level)

        return self.send("set_bright", [level])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting color temperature to {level}"),
    )
    def set_color_temperature(self, level: int):
        """Set Correlated Color Temperature."""
        if level < 1 or level > 100:
            raise PhilipsMoonlightException("Invalid color temperature: %s" % level)

        return self.send("set_cct", [level])

    @command(
        click.argument("brightness", type=int),
        click.argument("cct", type=int),
        default_output=format_output(
            "Setting brightness to {brightness} and color temperature to {cct}"
        ),
    )
    def set_brightness_and_color_temperature(self, brightness: int, cct: int):
        """Set brightness level and the correlated color temperature."""
        if brightness < 1 or brightness > 100:
            raise PhilipsMoonlightException("Invalid brightness: %s" % brightness)

        if cct < 1 or cct > 100:
            raise PhilipsMoonlightException("Invalid color temperature: %s" % cct)

        return self.send("set_bricct", [brightness, cct])

    @command(
        click.argument("brightness", type=int),
        click.argument("rgb", default=[255] * 3, type=click.Tuple([int, int, int])),
        default_output=format_output(
            "Setting brightness to {brightness} and color to {rgb}"
        ),
    )
    def set_brightness_and_rgb(self, brightness: int, rgb: Tuple[int, int, int]):
        """Set brightness level and the color."""
        if brightness < 1 or brightness > 100:
            raise PhilipsMoonlightException("Invalid brightness: %s" % brightness)

        for color in rgb:
            if color < 0 or color > 255:
                raise PhilipsMoonlightException("Invalid color: %s" % color)

        return self.send("set_brirgb", [*rgb, brightness])

    @command(
        click.argument("number", type=int),
        default_output=format_output("Setting fixed scene to {number}"),
    )
    def set_scene(self, number: int):
        """Set scene number."""
        if number < 1 or number > 6:
            raise PhilipsMoonlightException("Invalid fixed scene number: %s" % number)

        if number == 6:
            return self.send("go_night")

        return self.send("apply_fixed_scene", [number])



import base64
import re
from typing import Callable, Set, Tuple

import click
from construct import (
    Adapter,
    Array,
    BitsInteger,
    BitStruct,
    Computed,
    Const,
    Int16ub,
    Int16ul,
    Int32ul,
    Rebuild,
    Struct,
    len_,
    this,
)

from .click_common import command, format_output
from .device import Device
from .exceptions import DeviceException


class ChuangmiIrException(DeviceException):
    pass


class ChuangmiIr(Device):
    """Main class representing Chuangmi IR Remote Controller."""

    PRONTO_RE = re.compile(r"^([\da-f]{4}\s?){3,}([\da-f]{4})$", re.IGNORECASE)

    @command(
        click.argument("key", type=int),
        default_output=format_output("Learning command into storage key {key}"),
    )
    def learn(self, key: int = 1):
        """Learn an infrared command.

        :param int key: Storage slot, must be between 1 and 1000000
        """

        if key < 1 or key > 1000000:
            raise ChuangmiIrException("Invalid storage slot.")
        return self.send("miIO.ir_learn", {"key": str(key)})

    @command(
        click.argument("key", type=int),
        default_output=format_output("Reading infrared command from storage key {key}"),
    )
    def read(self, key: int = 1):
        """Read a learned command.

        Positive response (chuangmi.ir.v2):
        {'key': '1', 'code': 'Z6WPAasBAAA3BQAA4AwJAEA....AAABAAEBAQAAAQAA=='}

        Negative response (chuangmi.ir.v2):
        {'error': {'code': -5002, 'message': 'no code for this key'}, 'id': 5}

        Negative response (chuangmi.ir.v2):
        {'error': {'code': -5003, 'message': 'learn timeout'}, 'id': 17}

        :param int key: Slot to read from
        """

        if key < 1 or key > 1000000:
            raise ChuangmiIrException("Invalid storage slot.")
        return self.send("miIO.ir_read", {"key": str(key)})

    def play_raw(self, command: str, frequency: int = 38400, length: int = -1):
        """Play a captured command.

        :param str command: Command to execute
        :param int frequency: Execution frequency
        :param int length: Length of the command. -1 means not sending the length parameter.
        """
        if length < 0:
            return self.send("miIO.ir_play", {"freq": frequency, "code": command})
        else:
            return self.send(
                "miIO.ir_play", {"freq": frequency, "code": command, "length": length}
            )

    def play_pronto(self, pronto: str, repeats: int = 1, length: int = -1):
        """Play a Pronto Hex encoded IR command. Supports only raw Pronto format,
        starting with 0000.

        :param str pronto: Pronto Hex string.
        :param int repeats: Number of extra signal repeats.
        :param int length: Length of the command. -1 means not sending the length parameter.
        """
        command, frequency = self.pronto_to_raw(pronto, repeats)
        return self.play_raw(command, frequency, length)

    @classmethod
    def pronto_to_raw(cls, pronto: str, repeats: int = 1) -> Tuple[str, int]:
        """Play a Pronto Hex encoded IR command. Supports only raw Pronto format,
        starting with 0000.

        :param str pronto: Pronto Hex string.
        :param int repeats: Number of extra signal repeats.
        """
        if repeats < 0:
            raise ChuangmiIrException("Invalid repeats value")

        try:
            pronto_data = Pronto.parse(bytearray.fromhex(pronto))
        except Exception as ex:
            raise ChuangmiIrException("Invalid Pronto command") from ex

        if len(pronto_data.intro) == 0:
            repeats += 1

        times: Set[int] = set()
        for pair in pronto_data.intro + pronto_data.repeat * (1 if repeats else 0):
            times.add(pair.pulse)
            times.add(pair.gap)

        times_sorted = sorted(times)
        times_map = {t: idx for idx, t in enumerate(times_sorted)}
        edge_pairs = []
        for pair in pronto_data.intro + pronto_data.repeat * repeats:
            edge_pairs.append(
                {"pulse": times_map[pair.pulse], "gap": times_map[pair.gap]}
            )

        signal_code = base64.b64encode(
            ChuangmiIrSignal.build(
                {
                    "times_index": times_sorted + [0] * (16 - len(times)),
                    "edge_pairs": edge_pairs,
                }
            )
        ).decode()

        return signal_code, int(round(pronto_data.frequency))

    @command(
        click.argument("command", type=str),
        default_output=format_output("Playing the supplied command"),
    )
    def play(self, command: str):
        """Plays a command in one of the supported formats."""
        if ":" not in command:
            if self.PRONTO_RE.match(command):
                command_type = "pronto"
            else:
                command_type = "raw"
            command_args = []
        else:
            command_type, command, *command_args = command.split(":")

        arg_types = [int, int]
        if len(command_args) > len(arg_types):
            raise ChuangmiIrException("Invalid command arguments count")

        if command_type not in ["raw", "pronto"]:
            raise ChuangmiIrException("Invalid command type")

        play_method: Callable
        if command_type == "raw":
            play_method = self.play_raw

        elif command_type == "pronto":
            play_method = self.play_pronto

        try:
            converted_command_args = [t(v) for v, t in zip(command_args, arg_types)]
        except Exception as ex:
            raise ChuangmiIrException("Invalid command arguments") from ex

        return play_method(command, *converted_command_args)

    @command(
        click.argument("indicator_led", type=bool),
        default_output=format_output(
            lambda indicator_led: "Turning on indicator LED"
            if indicator_led
            else "Turning off indicator LED"
        ),
    )
    def set_indicator_led(self, indicator_led: bool):
        """Set the indicator led on/off."""
        if indicator_led:
            return self.send("set_indicatorLamp", ["on"])
        else:
            return self.send("set_indicatorLamp", ["off"])

    @command(default_output=format_output("Indicator LED status: {result}"))
    def get_indicator_led(self):
        """Get the indicator led status."""
        return self.send("get_indicatorLamp")


class ProntoPulseAdapter(Adapter):
    def _decode(self, obj, context, path):
        return int(obj * context._.modulation_period)

    def _encode(self, obj, context, path):
        raise RuntimeError("Not implemented")


ChuangmiIrSignal = Struct(
    Const(0xA567, Int16ul),
    "edge_count" / Rebuild(Int16ul, len_(this.edge_pairs) * 2 - 1),
    "times_index" / Array(16, Int32ul),
    "edge_pairs"
    / Array(
        (this.edge_count + 1) // 2,
        BitStruct("gap" / BitsInteger(4), "pulse" / BitsInteger(4)),
    ),
)

ProntoBurstPair = Struct(
    "pulse" / ProntoPulseAdapter(Int16ub), "gap" / ProntoPulseAdapter(Int16ub)
)

Pronto = Struct(
    Const(0, Int16ub),
    "_ticks" / Int16ub,
    "modulation_period" / Computed(this._ticks * 0.241246),
    "frequency" / Computed(1000000 / this.modulation_period),
    "intro_len" / Int16ub,
    "repeat_len" / Int16ub,
    "intro" / Array(this.intro_len, ProntoBurstPair),
    "repeat" / Array(this.repeat_len, ProntoBurstPair),
)



"""Xiaomi Gateway Radio implementation."""

import click

from ..click_common import command
from .gatewaydevice import GatewayDevice


class Radio(GatewayDevice):
    """Radio controls for the gateway."""

    @command()
    def get_radio_info(self):
        """Radio play info."""
        return self._gateway.send("get_prop_fm")

    @command(click.argument("volume"))
    def set_radio_volume(self, volume):
        """Set radio volume."""
        return self._gateway.send("set_fm_volume", [volume])

    def play_music_new(self):
        """Unknown."""
        # {'from': '4', 'id': 9514,
        #  'method': 'set_default_music', 'params': [2, '21']}
        # {'from': '4', 'id': 9515,
        #  'method': 'play_music_new', 'params': ['21', 0]}
        raise NotImplementedError()

    def play_specify_fm(self):
        """play specific stream?"""
        raise NotImplementedError()
        # {"from": "4", "id": 65055, "method": "play_specify_fm",
        # "params": {"id": 764, "type": 0,
        # "url": "http://live.xmcdn.com/live/764/64.m3u8"}}
        return self._gateway.send("play_specify_fm")

    def play_fm(self):
        """radio on/off?"""
        raise NotImplementedError()
        # play_fm","params":["off"]}
        return self._gateway.send("play_fm")

    def volume_ctrl_fm(self):
        """Unknown."""
        raise NotImplementedError()
        return self._gateway.send("volume_ctrl_fm")

    def get_channels(self):
        """Unknown."""
        raise NotImplementedError()
        # "method": "get_channels", "params": {"start": 0}}
        return self._gateway.send("get_channels")

    def add_channels(self):
        """Unknown."""
        raise NotImplementedError()
        return self._gateway.send("add_channels")

    def remove_channels(self):
        """Unknown."""
        raise NotImplementedError()
        return self._gateway.send("remove_channels")

    def get_default_music(self):
        """seems to timeout (w/o internet)."""
        # params [0,1,2]
        raise NotImplementedError()
        return self._gateway.send("get_default_music")

    @command()
    def get_music_info(self):
        """Unknown."""
        info = self._gateway.send("get_music_info")
        click.echo("info: %s" % info)
        free_space = self._gateway.send("get_music_free_space")
        click.echo("free space: %s" % free_space)

    @command()
    def get_mute(self):
        """mute of what?"""
        return self._gateway.send("get_mute")

    def download_music(self):
        """Unknown."""
        raise NotImplementedError()
        return self._gateway.send("download_music")

    def delete_music(self):
        """delete music."""
        raise NotImplementedError()
        return self._gateway.send("delete_music")

    def download_user_music(self):
        """Unknown."""
        raise NotImplementedError()
        return self._gateway.send("download_user_music")

    def get_download_progress(self):
        """progress for music downloads or updates?"""
        # returns [':0']
        raise NotImplementedError()
        return self._gateway.send("get_download_progress")

    @command()
    def set_sound_playing(self):
        """stop playing?"""
        return self._gateway.send("set_sound_playing", ["off"])

    def set_default_music(self):
        """Unknown."""
        raise NotImplementedError()
        # method":"set_default_music","params":[0,"2"]}



from unittest import TestCase

import pytest

from miio import Fan1C, FanMiot
from miio.fan_miot import (
    MODEL_FAN_1C,
    MODEL_FAN_P9,
    MODEL_FAN_P10,
    MODEL_FAN_P11,
    FanException,
    OperationMode,
)

from .dummies import DummyMiotDevice


class DummyFanMiot(DummyMiotDevice, FanMiot):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_P9
        self.state = {
            "power": True,
            "mode": 0,
            "fan_speed": 35,
            "swing_mode": False,
            "swing_mode_angle": 30,
            "power_off_time": 0,
            "light": True,
            "buzzer": False,
            "child_lock": False,
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fanmiot(request):
    request.cls.device = DummyFanMiot()


@pytest.mark.usefixtures("fanmiot")
class TestFanMiot(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Normal)
        assert mode() == OperationMode.Normal

        self.device.set_mode(OperationMode.Nature)
        assert mode() == OperationMode.Nature

    def test_set_speed(self):
        def speed():
            return self.device.status().speed

        self.device.set_speed(0)
        assert speed() == 0
        self.device.set_speed(1)
        assert speed() == 1
        self.device.set_speed(100)
        assert speed() == 100

        with pytest.raises(FanException):
            self.device.set_speed(-1)

        with pytest.raises(FanException):
            self.device.set_speed(101)

    def test_set_angle(self):
        def angle():
            return self.device.status().angle

        self.device.set_angle(30)
        assert angle() == 30
        self.device.set_angle(60)
        assert angle() == 60
        self.device.set_angle(90)
        assert angle() == 90
        self.device.set_angle(120)
        assert angle() == 120
        self.device.set_angle(150)
        assert angle() == 150

        with pytest.raises(FanException):
            self.device.set_angle(-1)

        with pytest.raises(FanException):
            self.device.set_angle(1)

        with pytest.raises(FanException):
            self.device.set_angle(31)

        with pytest.raises(FanException):
            self.device.set_angle(140)

        with pytest.raises(FanException):
            self.device.set_angle(151)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(0)
        assert delay_off_countdown() == 0
        self.device.delay_off(1)
        assert delay_off_countdown() == 1
        self.device.delay_off(480)
        assert delay_off_countdown() == 480

        with pytest.raises(FanException):
            self.device.delay_off(-1)
        with pytest.raises(FanException):
            self.device.delay_off(481)


class DummyFanMiotP10(DummyFanMiot, FanMiot):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)
        self.model = MODEL_FAN_P10


@pytest.fixture(scope="class")
def fanmiotp10(request):
    request.cls.device = DummyFanMiotP10()


@pytest.mark.usefixtures("fanmiotp10")
class TestFanMiotP10(TestCase):
    def test_set_angle(self):
        def angle():
            return self.device.status().angle

        self.device.set_angle(30)
        assert angle() == 30
        self.device.set_angle(60)
        assert angle() == 60
        self.device.set_angle(90)
        assert angle() == 90
        self.device.set_angle(120)
        assert angle() == 120
        self.device.set_angle(140)
        assert angle() == 140

        with pytest.raises(FanException):
            self.device.set_angle(-1)

        with pytest.raises(FanException):
            self.device.set_angle(1)

        with pytest.raises(FanException):
            self.device.set_angle(31)

        with pytest.raises(FanException):
            self.device.set_angle(150)

        with pytest.raises(FanException):
            self.device.set_angle(141)


class DummyFanMiotP11(DummyFanMiot, FanMiot):
    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)
        self.model = MODEL_FAN_P11


@pytest.fixture(scope="class")
def fanmiotp11(request):
    request.cls.device = DummyFanMiotP11()


@pytest.mark.usefixtures("fanmiotp11")
class TestFanMiotP11(TestFanMiotP10, TestCase):
    pass


class DummyFan1C(DummyMiotDevice, Fan1C):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_1C
        self.state = {
            "power": True,
            "mode": 0,
            "fan_level": 1,
            "swing_mode": False,
            "power_off_time": 0,
            "light": True,
            "buzzer": False,
            "child_lock": False,
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fan1c(request):
    request.cls.device = DummyFan1C()


@pytest.mark.usefixtures("fan1c")
class TestFan1C(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Normal)
        assert mode() == OperationMode.Normal

        self.device.set_mode(OperationMode.Nature)
        assert mode() == OperationMode.Nature

    def test_set_speed(self):
        def speed():
            return self.device.status().speed

        self.device.set_speed(1)
        assert speed() == 1
        self.device.set_speed(2)
        assert speed() == 2
        self.device.set_speed(3)
        assert speed() == 3

        with pytest.raises(FanException):
            self.device.set_speed(0)

        with pytest.raises(FanException):
            self.device.set_speed(4)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(0)
        assert delay_off_countdown() == 0
        self.device.delay_off(1)
        assert delay_off_countdown() == 1
        self.device.delay_off(480)
        assert delay_off_countdown() == 480

        with pytest.raises(FanException):
            self.device.delay_off(-1)
        with pytest.raises(FanException):
            self.device.delay_off(481)



from unittest import TestCase

import pytest

from miio import Fan, FanP5
from miio.fan import (
    MODEL_FAN_P5,
    MODEL_FAN_SA1,
    MODEL_FAN_V2,
    MODEL_FAN_V3,
    FanException,
    FanStatus,
    FanStatusP5,
    LedBrightness,
    MoveDirection,
    OperationMode,
)

from .dummies import DummyDevice


class DummyFanV2(DummyDevice, Fan):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_V2
        # This example response is just a guess. Please update!
        self.state = {
            "temp_dec": 232,
            "humidity": 46,
            "angle": 118,
            "speed": 298,
            "poweroff_time": 0,
            "power": "on",
            "ac_power": "off",
            "battery": 98,
            "angle_enable": "off",
            "speed_level": 1,
            "natural_level": 0,
            "child_lock": "off",
            "buzzer": "on",
            "led_b": 1,
            "led": "on",
            "natural_enable": None,
            "use_time": 0,
            "bat_charge": "complete",
            "bat_state": None,
            "button_pressed": "speed",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_speed_level": lambda x: self._set_state("speed_level", x),
            "set_natural_level": lambda x: self._set_state("natural_level", x),
            "set_move": lambda x: True,
            "set_angle": lambda x: self._set_state("angle", x),
            "set_angle_enable": lambda x: self._set_state("angle_enable", x),
            "set_led_b": lambda x: self._set_state("led_b", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_poweroff_time": lambda x: self._set_state("poweroff_time", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fanv2(request):
    request.cls.device = DummyFanV2()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("fanv2")
class TestFanV2(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(FanStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temp_dec"] / 10.0
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().angle == self.device.start_state["angle"]
        assert self.state().speed == self.device.start_state["speed"]
        assert (
            self.state().delay_off_countdown == self.device.start_state["poweroff_time"]
        )
        assert self.state().ac_power is (self.device.start_state["ac_power"] == "on")
        assert self.state().battery == self.device.start_state["battery"]
        assert self.state().oscillate is (
            self.device.start_state["angle_enable"] == "on"
        )
        assert self.state().direct_speed == self.device.start_state["speed_level"]
        assert self.state().natural_speed == self.device.start_state["natural_level"]
        assert self.state().child_lock is (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().buzzer is (self.device.start_state["buzzer"] == "on")
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_b"]
        )
        assert self.state().led is (self.device.start_state["led"] == "on")
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().battery_charge == self.device.start_state["bat_charge"]
        assert self.state().battery_state == self.device.start_state["bat_state"]
        assert self.state().button_pressed == self.device.start_state["button_pressed"]

    def test_status_without_led_brightness(self):
        self.device._reset_state()

        self.device.state["led_b"] = None
        assert self.state().led_brightness is None

    def test_status_without_battery_charge(self):
        self.device._reset_state()

        self.device.state["bat_charge"] = None
        assert self.state().battery_charge is None

    def test_status_without_battery_state(self):
        self.device._reset_state()

        self.device.state["bat_state"] = None
        assert self.state().battery_state is None

    def test_status_without_button_pressed(self):
        self.device._reset_state()

        self.device.state["button_pressed"] = None
        assert self.state().button_pressed is None

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_direct_speed(self):
        def direct_speed():
            return self.device.status().direct_speed

        self.device.set_direct_speed(0)
        assert direct_speed() == 0
        self.device.set_direct_speed(1)
        assert direct_speed() == 1
        self.device.set_direct_speed(100)
        assert direct_speed() == 100

        with pytest.raises(FanException):
            self.device.set_direct_speed(-1)

        with pytest.raises(FanException):
            self.device.set_direct_speed(101)

    def test_set_rotate(self):
        """The method is open-loop.

        The new state cannot be retrieved.
        """
        self.device.set_rotate(MoveDirection.Left)
        self.device.set_rotate(MoveDirection.Right)

    def test_set_angle(self):
        """This test doesn't implement the real behaviour of the device may be.

        The property "angle" doesn't provide the current setting. It's a measurement of
        the current position probably.
        """

        def angle():
            return self.device.status().angle

        self.device.set_angle(0)  # TODO: Is this value allowed?
        assert angle() == 0
        self.device.set_angle(1)  # TODO: Is this value allowed?
        assert angle() == 1
        self.device.set_angle(30)
        assert angle() == 30
        self.device.set_angle(60)
        assert angle() == 60
        self.device.set_angle(90)
        assert angle() == 90
        self.device.set_angle(120)
        assert angle() == 120

        with pytest.raises(FanException):
            self.device.set_angle(-1)

        with pytest.raises(FanException):
            self.device.set_angle(121)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200
        self.device.delay_off(0)
        assert delay_off_countdown() == 0

        with pytest.raises(FanException):
            self.device.delay_off(-1)


class DummyFanV3(DummyDevice, Fan):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_V3
        self.state = {
            "temp_dec": 232,
            "humidity": 46,
            "angle": 118,
            "speed": 298,
            "poweroff_time": 0,
            "power": "on",
            "ac_power": "off",
            "battery": 98,
            "angle_enable": "off",
            "speed_level": 1,
            "natural_level": 0,
            "child_lock": "off",
            "buzzer": "on",
            "led_b": 1,
            "led": None,
            "natural_enable": None,
            "use_time": 0,
            "bat_charge": "complete",
            "bat_state": None,
            "button_pressed": "speed",
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_speed_level": lambda x: self._set_state("speed_level", x),
            "set_natural_level": lambda x: self._set_state("natural_level", x),
            "set_move": lambda x: True,
            "set_angle": lambda x: self._set_state("angle", x),
            "set_angle_enable": lambda x: self._set_state("angle_enable", x),
            "set_led_b": lambda x: self._set_state("led_b", x),
            "set_led": lambda x: self._set_state("led", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_poweroff_time": lambda x: self._set_state("poweroff_time", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fanv3(request):
    request.cls.device = DummyFanV3()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("fanv3")
class TestFanV3(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(FanStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().temperature == self.device.start_state["temp_dec"] / 10.0
        assert self.state().humidity == self.device.start_state["humidity"]
        assert self.state().angle == self.device.start_state["angle"]
        assert self.state().speed == self.device.start_state["speed"]
        assert (
            self.state().delay_off_countdown == self.device.start_state["poweroff_time"]
        )
        assert self.state().ac_power is (self.device.start_state["ac_power"] == "on")
        assert self.state().battery == self.device.start_state["battery"]
        assert self.state().oscillate is (
            self.device.start_state["angle_enable"] == "on"
        )
        assert self.state().direct_speed == self.device.start_state["speed_level"]
        assert self.state().natural_speed == self.device.start_state["natural_level"]
        assert self.state().child_lock is (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().buzzer is (self.device.start_state["buzzer"] == "on")
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_b"]
        )
        assert self.state().led is None
        assert self.state().use_time == self.device.start_state["use_time"]
        assert self.state().battery_charge == self.device.start_state["bat_charge"]
        assert self.state().battery_state == self.device.start_state["bat_state"]
        assert self.state().button_pressed == self.device.start_state["button_pressed"]

    def test_status_without_led_brightness(self):
        self.device._reset_state()

        self.device.state["led_b"] = None
        assert self.state().led_brightness is None

    def test_status_without_battery_charge(self):
        self.device._reset_state()

        self.device.state["bat_charge"] = None
        assert self.state().battery_charge is None

    def test_status_without_battery_state(self):
        self.device._reset_state()

        self.device.state["bat_state"] = None
        assert self.state().battery_state is None

    def test_status_without_button_pressed(self):
        self.device._reset_state()

        self.device.state["button_pressed"] = None
        assert self.state().button_pressed is None

    def test_set_direct_speed(self):
        def direct_speed():
            return self.device.status().direct_speed

        self.device.set_direct_speed(0)
        assert direct_speed() == 0
        self.device.set_direct_speed(1)
        assert direct_speed() == 1
        self.device.set_direct_speed(100)
        assert direct_speed() == 100

        with pytest.raises(FanException):
            self.device.set_direct_speed(-1)

        with pytest.raises(FanException):
            self.device.set_direct_speed(101)

    def test_set_natural_speed(self):
        def natural_speed():
            return self.device.status().natural_speed

        self.device.set_natural_speed(0)
        assert natural_speed() == 0
        self.device.set_natural_speed(1)
        assert natural_speed() == 1
        self.device.set_natural_speed(100)
        assert natural_speed() == 100

        with pytest.raises(FanException):
            self.device.set_natural_speed(-1)

        with pytest.raises(FanException):
            self.device.set_natural_speed(101)

    def test_set_rotate(self):
        """The method is open-loop.

        The new state cannot be retrieved.
        """
        self.device.set_rotate(MoveDirection.Left)
        self.device.set_rotate(MoveDirection.Right)

    def test_set_angle(self):
        """This test doesn't implement the real behaviour of the device may be.

        The property "angle" doesn't provide the current setting. It's a measurement of
        the current position probably.
        """

        def angle():
            return self.device.status().angle

        self.device.set_angle(0)  # TODO: Is this value allowed?
        assert angle() == 0
        self.device.set_angle(1)  # TODO: Is this value allowed?
        assert angle() == 1
        self.device.set_angle(30)
        assert angle() == 30
        self.device.set_angle(60)
        assert angle() == 60
        self.device.set_angle(90)
        assert angle() == 90
        self.device.set_angle(120)
        assert angle() == 120

        with pytest.raises(FanException):
            self.device.set_angle(-1)

        with pytest.raises(FanException):
            self.device.set_angle(121)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200
        self.device.delay_off(0)
        assert delay_off_countdown() == 0

        with pytest.raises(FanException):
            self.device.delay_off(-1)


class DummyFanSA1(DummyDevice, Fan):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_SA1
        self.state = {
            "angle": 120,
            "speed": 277,
            "poweroff_time": 0,
            "power": "on",
            "ac_power": "on",
            "angle_enable": "off",
            "speed_level": 1,
            "natural_level": 2,
            "child_lock": "off",
            "buzzer": 0,
            "led_b": 0,
            "use_time": 2318,
        }

        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_speed_level": lambda x: self._set_state("speed_level", x),
            "set_natural_level": lambda x: self._set_state("natural_level", x),
            "set_move": lambda x: True,
            "set_angle": lambda x: self._set_state("angle", x),
            "set_angle_enable": lambda x: self._set_state("angle_enable", x),
            "set_led_b": lambda x: self._set_state("led_b", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_poweroff_time": lambda x: self._set_state("poweroff_time", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fansa1(request):
    request.cls.device = DummyFanSA1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("fansa1")
class TestFanSA1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(FanStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().angle == self.device.start_state["angle"]
        assert self.state().speed == self.device.start_state["speed"]
        assert (
            self.state().delay_off_countdown == self.device.start_state["poweroff_time"]
        )
        assert self.state().ac_power is (self.device.start_state["ac_power"] == "on")
        assert self.state().oscillate is (
            self.device.start_state["angle_enable"] == "on"
        )
        assert self.state().direct_speed == self.device.start_state["speed_level"]
        assert self.state().natural_speed == self.device.start_state["natural_level"]
        assert self.state().child_lock is (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().buzzer is (self.device.start_state["buzzer"] == "on")
        assert self.state().led_brightness == LedBrightness(
            self.device.start_state["led_b"]
        )
        assert self.state().led is None
        assert self.state().use_time == self.device.start_state["use_time"]

    def test_set_direct_speed(self):
        def direct_speed():
            return self.device.status().direct_speed

        self.device.set_direct_speed(0)
        assert direct_speed() == 0
        self.device.set_direct_speed(1)
        assert direct_speed() == 1
        self.device.set_direct_speed(100)
        assert direct_speed() == 100

        with pytest.raises(FanException):
            self.device.set_direct_speed(-1)

        with pytest.raises(FanException):
            self.device.set_direct_speed(101)

    def test_set_natural_speed(self):
        def natural_speed():
            return self.device.status().natural_speed

        self.device.set_natural_speed(0)
        assert natural_speed() == 0
        self.device.set_natural_speed(1)
        assert natural_speed() == 1
        self.device.set_natural_speed(100)
        assert natural_speed() == 100

        with pytest.raises(FanException):
            self.device.set_natural_speed(-1)

        with pytest.raises(FanException):
            self.device.set_natural_speed(101)

    def test_set_rotate(self):
        """The method is open-loop.

        The new state cannot be retrieved.
        """
        self.device.set_rotate(MoveDirection.Left)
        self.device.set_rotate(MoveDirection.Right)

    def test_set_angle(self):
        """This test doesn't implement the real behaviour of the device may be.

        The property "angle" doesn't provide the current setting. It's a measurement of
        the current position probably.
        """

        def angle():
            return self.device.status().angle

        self.device.set_angle(0)  # TODO: Is this value allowed?
        assert angle() == 0
        self.device.set_angle(1)  # TODO: Is this value allowed?
        assert angle() == 1
        self.device.set_angle(30)
        assert angle() == 30
        self.device.set_angle(60)
        assert angle() == 60
        self.device.set_angle(90)
        assert angle() == 90
        self.device.set_angle(120)
        assert angle() == 120

        with pytest.raises(FanException):
            self.device.set_angle(-1)

        with pytest.raises(FanException):
            self.device.set_angle(121)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_led_brightness(self):
        def led_brightness():
            return self.device.status().led_brightness

        self.device.set_led_brightness(LedBrightness.Bright)
        assert led_brightness() == LedBrightness.Bright

        self.device.set_led_brightness(LedBrightness.Dim)
        assert led_brightness() == LedBrightness.Dim

        self.device.set_led_brightness(LedBrightness.Off)
        assert led_brightness() == LedBrightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200
        self.device.delay_off(0)
        assert delay_off_countdown() == 0

        with pytest.raises(FanException):
            self.device.delay_off(-1)


class DummyFanP5(DummyDevice, FanP5):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_P5
        self.state = {
            "power": True,
            "mode": "normal",
            "speed": 35,
            "roll_enable": False,
            "roll_angle": 140,
            "time_off": 0,
            "light": True,
            "beep_sound": False,
            "child_lock": False,
        }

        self.return_values = {
            "get_prop": self._get_state,
            "s_power": lambda x: self._set_state("power", x),
            "s_mode": lambda x: self._set_state("mode", x),
            "s_speed": lambda x: self._set_state("speed", x),
            "s_roll": lambda x: self._set_state("roll_enable", x),
            "s_angle": lambda x: self._set_state("roll_angle", x),
            "s_t_off": lambda x: self._set_state("time_off", x),
            "s_light": lambda x: self._set_state("light", x),
            "s_sound": lambda x: self._set_state("beep_sound", x),
            "s_lock": lambda x: self._set_state("child_lock", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fanp5(request):
    request.cls.device = DummyFanP5()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("fanp5")
class TestFanP5(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(FanStatusP5(self.device.start_state))

        assert self.is_on() is True
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().speed == self.device.start_state["speed"]
        assert self.state().oscillate is self.device.start_state["roll_enable"]
        assert self.state().angle == self.device.start_state["roll_angle"]
        assert self.state().delay_off_countdown == self.device.start_state["time_off"]
        assert self.state().led is self.device.start_state["light"]
        assert self.state().buzzer is self.device.start_state["beep_sound"]
        assert self.state().child_lock is self.device.start_state["child_lock"]

    def test_set_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_mode(OperationMode.Normal)
        assert mode() == OperationMode.Normal

        self.device.set_mode(OperationMode.Nature)
        assert mode() == OperationMode.Nature

    def test_set_speed(self):
        def speed():
            return self.device.status().speed

        self.device.set_speed(0)
        assert speed() == 0
        self.device.set_speed(1)
        assert speed() == 1
        self.device.set_speed(100)
        assert speed() == 100

        with pytest.raises(FanException):
            self.device.set_speed(-1)

        with pytest.raises(FanException):
            self.device.set_speed(101)

    def test_set_angle(self):
        def angle():
            return self.device.status().angle

        self.device.set_angle(30)
        assert angle() == 30
        self.device.set_angle(60)
        assert angle() == 60
        self.device.set_angle(90)
        assert angle() == 90
        self.device.set_angle(120)
        assert angle() == 120
        self.device.set_angle(140)
        assert angle() == 140

        with pytest.raises(FanException):
            self.device.set_angle(-1)

        with pytest.raises(FanException):
            self.device.set_angle(1)

        with pytest.raises(FanException):
            self.device.set_angle(31)

        with pytest.raises(FanException):
            self.device.set_angle(141)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_led(self):
        def led():
            return self.device.status().led

        self.device.set_led(True)
        assert led() is True

        self.device.set_led(False)
        assert led() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200
        self.device.delay_off(0)
        assert delay_off_countdown() == 0

        with pytest.raises(FanException):
            self.device.delay_off(-1)



import inspect
import logging
from enum import Enum
from pprint import pformat as pf
from typing import Any, Optional  # noqa: F401

import click

from .click_common import DeviceGroupMeta, LiteralParamType, command, format_output
from .deviceinfo import DeviceInfo
from .exceptions import DeviceInfoUnavailableException, PayloadDecodeException
from .miioprotocol import MiIOProtocol

_LOGGER = logging.getLogger(__name__)


class UpdateState(Enum):
    Downloading = "downloading"
    Installing = "installing"
    Failed = "failed"
    Idle = "idle"


class DeviceStatus:
    """Base class for status containers.

    All status container classes should inherit from this class. The __repr__
    implementation returns all defined properties and their values.
    """

    def __repr__(self):
        props = inspect.getmembers(self.__class__, lambda o: isinstance(o, property))

        s = f"<{self.__class__.__name__}"
        for prop_tuple in props:
            name, prop = prop_tuple
            try:
                prop_value = prop.fget(self)
            except Exception as ex:
                prop_value = ex.__class__.__name__

            s += f" {name}={prop_value}"
        s += ">"
        return s


class Device(metaclass=DeviceGroupMeta):
    """Base class for all device implementations.

    This is the main class providing the basic protocol handling for devices using the
    ``miIO`` protocol. This class should not be initialized directly but a device-
    specific class inheriting it should be used instead of it.
    """

    retry_count = 3
    timeout = 5

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        timeout: int = None,
    ) -> None:
        self.ip = ip
        self.token = token
        timeout = timeout if timeout is not None else self.timeout
        self._protocol = MiIOProtocol(
            ip, token, start_id, debug, lazy_discover, timeout
        )

    def send(
        self,
        command: str,
        parameters: Any = None,
        retry_count: int = None,
        *,
        extra_parameters=None,
    ) -> Any:
        """Send a command to the device.

        Basic format of the request:
        {"id": 1234, "method": command, "parameters": parameters}

        `extra_parameters` allows passing elements to the top-level of the request.
        This is necessary for some devices, such as gateway devices, which expect
        the sub-device identifier to be on the top-level.

        :param str command: Command to send
        :param dict parameters: Parameters to send
        :param int retry_count: How many times to retry on error
        :param dict extra_parameters: Extra top-level parameters
        """
        retry_count = retry_count if retry_count is not None else self.retry_count
        return self._protocol.send(
            command, parameters, retry_count, extra_parameters=extra_parameters
        )

    def send_handshake(self):
        """Send initial handshake to the device."""
        return self._protocol.send_handshake()

    @command(
        click.argument("command", type=str, required=True),
        click.argument("parameters", type=LiteralParamType(), required=False),
    )
    def raw_command(self, command, parameters):
        """Send a raw command to the device. This is mostly useful when trying out
        commands which are not implemented by a given device instance.

        :param str command: Command to send
        :param dict parameters: Parameters to send
        """
        return self.send(command, parameters)

    @command(
        default_output=format_output(
            "",
            "Model: {result.model}\n"
            "Hardware version: {result.hardware_version}\n"
            "Firmware version: {result.firmware_version}\n",
        )
    )
    def info(self) -> DeviceInfo:
        """Get miIO protocol information from the device.

        This includes information about connected wlan network, and hardware and
        software versions.
        """
        try:
            return DeviceInfo(self.send("miIO.info"))
        except PayloadDecodeException as ex:
            raise DeviceInfoUnavailableException(
                "Unable to request miIO.info from the device"
            ) from ex

    @property
    def raw_id(self):
        """Return the last used protocol sequence id."""
        return self._protocol.raw_id

    def update(self, url: str, md5: str):
        """Start an OTA update."""
        payload = {
            "mode": "normal",
            "install": "1",
            "app_url": url,
            "file_md5": md5,
            "proc": "dnld install",
        }
        return self.send("miIO.ota", payload)[0] == "ok"

    def update_progress(self) -> int:
        """Return current update progress [0-100]."""
        return self.send("miIO.get_ota_progress")[0]

    def update_state(self):
        """Return current update state."""
        return UpdateState(self.send("miIO.get_ota_state")[0])

    def configure_wifi(self, ssid, password, uid=0, extra_params=None):
        """Configure the wifi settings."""
        if extra_params is None:
            extra_params = {}
        params = {"ssid": ssid, "passwd": password, "uid": uid, **extra_params}

        return self.send("miIO.config_router", params)[0]

    def get_properties(
        self, properties, *, property_getter="get_prop", max_properties=None
    ):
        """Request properties in slices based on given max_properties.

        This is necessary as some devices have limitation on how many
        properties can be queried at once.

        If `max_properties` is None, all properties are requested at once.

        :param list properties: List of properties to query from the device.
        :param int max_properties: Number of properties that can be requested at once.
        :return List of property values.
        """
        _props = properties.copy()
        values = []
        while _props:
            values.extend(self.send(property_getter, _props[:max_properties]))
            if max_properties is None:
                break

            _props[:] = _props[max_properties:]

        properties_count = len(properties)
        values_count = len(values)
        if properties_count != values_count:
            _LOGGER.debug(
                "Count (%s) of requested properties does not match the "
                "count (%s) of received values.",
                properties_count,
                values_count,
            )

        return values

    @command(
        click.argument("properties", type=str, nargs=-1, required=True),
    )
    def test_properties(self, properties):
        """Helper to test device properties."""

        def ok(x):
            click.echo(click.style(x, fg="green", bold=True))

        def fail(x):
            click.echo(click.style(x, fg="red", bold=True))

        try:
            model = self.info().model
        except Exception as ex:
            _LOGGER.warning("Unable to obtain device model: %s", ex)
            model = "<unavailable>"

        click.echo(f"Testing properties {properties} for {model}")
        valid_properties = {}
        max_property_len = max([len(p) for p in properties])
        for property in properties:
            try:
                click.echo(f"Testing {property:{max_property_len+2}} ", nl=False)
                value = self.get_properties([property])
                # Handle list responses
                if isinstance(value, list):
                    # unwrap single-element lists
                    if len(value) == 1:
                        value = value.pop()
                    # report on unexpected multi-element lists
                    elif len(value) > 1:
                        _LOGGER.error("Got an array as response: %s", value)
                    # otherwise we received an empty list, which we consider here as None
                    else:
                        value = None

                if value is None:
                    fail("None")
                else:
                    valid_properties[property] = value
                    ok(f"{repr(value)} {type(value)}")
            except Exception as ex:
                _LOGGER.warning("Unable to request %s: %s", property, ex)

        click.echo(
            f"Found {len(valid_properties)} valid properties, testing max_properties.."
        )

        props_to_test = list(valid_properties.keys())
        max_properties = -1
        while len(props_to_test) > 1:
            try:
                click.echo(
                    f"Testing {len(props_to_test)} properties at once ({' '.join(props_to_test)}): ",
                    nl=False,
                )
                resp = self.get_properties(props_to_test)

                if len(resp) == len(props_to_test):
                    max_properties = len(props_to_test)
                    ok(f"OK for {max_properties} properties")
                    break
                else:
                    removed_property = props_to_test.pop()
                    fail(
                        f"Got different amount of properties ({len(props_to_test)}) than requested ({len(resp)}), removing {removed_property}"
                    )

            except Exception as ex:
                removed_property = props_to_test.pop()
                msg = f"Unable to request properties: {ex} - removing {removed_property} for next try"
                _LOGGER.warning(msg)
                fail(ex)

        non_empty_properties = {
            k: v for k, v in valid_properties.items() if v is not None
        }

        click.echo(
            click.style("\nPlease copy the results below to your report", bold=True)
        )
        click.echo("### Results ###")
        click.echo(f"Model: {model}")
        _LOGGER.debug(f"All responsive properties:\n{pf(valid_properties)}")
        click.echo(f"Total responsives: {len(valid_properties)}")
        click.echo(f"Total non-empty: {len(non_empty_properties)}")
        click.echo(f"All non-empty properties:\n{pf(non_empty_properties)}")
        click.echo(f"Max properties: {max_properties}")

        return "Done"

    def __repr__(self):
        return f"<{self.__class__.__name__ }: {self.ip} (token: {self.token})>"



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceInfo, DeviceStatus
from .exceptions import DeviceError, DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_HUMIDIFIER_V1 = "zhimi.humidifier.v1"
MODEL_HUMIDIFIER_CA1 = "zhimi.humidifier.ca1"
MODEL_HUMIDIFIER_CB1 = "zhimi.humidifier.cb1"
MODEL_HUMIDIFIER_CB2 = "zhimi.humidifier.cb2"

AVAILABLE_PROPERTIES_COMMON = [
    "power",
    "mode",
    "humidity",
    "buzzer",
    "led_b",
    "child_lock",
    "limit_hum",
    "use_time",
    "hw_version",
]

AVAILABLE_PROPERTIES = {
    MODEL_HUMIDIFIER_V1: AVAILABLE_PROPERTIES_COMMON
    + ["temp_dec", "trans_level", "button_pressed"],
    MODEL_HUMIDIFIER_CA1: AVAILABLE_PROPERTIES_COMMON
    + ["temp_dec", "speed", "depth", "dry"],
    MODEL_HUMIDIFIER_CB1: AVAILABLE_PROPERTIES_COMMON
    + ["temperature", "speed", "depth", "dry"],
    MODEL_HUMIDIFIER_CB2: AVAILABLE_PROPERTIES_COMMON
    + ["temperature", "speed", "depth", "dry"],
}


class AirHumidifierException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Silent = "silent"
    Medium = "medium"
    High = "high"
    Auto = "auto"
    Strong = "strong"


class LedBrightness(enum.Enum):
    Bright = 0
    Dim = 1
    Off = 2


class AirHumidifierStatus(DeviceStatus):
    """Container for status reports from the air humidifier."""

    def __init__(self, data: Dict[str, Any], device_info: DeviceInfo) -> None:
        """Response of a Air Humidifier (zhimi.humidifier.v1):

        {'power': 'off', 'mode': 'high', 'temp_dec': 294,
         'humidity': 33, 'buzzer': 'on', 'led_b': 0,
         'child_lock': 'on', 'limit_hum': 40, 'trans_level': 85,
         'speed': None, 'depth': None, 'dry': None, 'use_time': 941100,
         'hw_version': 0, 'button_pressed': 'led'}
        """

        self.data = data
        self.device_info = device_info

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if device is turned on."""
        return self.power == "on"

    @property
    def mode(self) -> OperationMode:
        """Operation mode.

        Can be either silent, medium or high.
        """
        return OperationMode(self.data["mode"])

    @property
    def temperature(self) -> Optional[float]:
        """Current temperature, if available."""
        if "temp_dec" in self.data and self.data["temp_dec"] is not None:
            return self.data["temp_dec"] / 10.0
        if "temperature" in self.data and self.data["temperature"] is not None:
            return self.data["temperature"]
        return None

    @property
    def humidity(self) -> int:
        """Current humidity."""
        return self.data["humidity"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["buzzer"] == "on"

    @property
    def led_brightness(self) -> Optional[LedBrightness]:
        """LED brightness if available."""
        if self.data["led_b"] is not None:
            return LedBrightness(self.data["led_b"])
        return None

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"] == "on"

    @property
    def target_humidity(self) -> int:
        """Target humidity.

        Can be either 30, 40, 50, 60, 70, 80 percent.
        """
        return self.data["limit_hum"]

    @property
    def trans_level(self) -> Optional[int]:
        """The meaning of the property is unknown.

        The property is used to determine the strong mode is enabled on old firmware.
        """
        if "trans_level" in self.data and self.data["trans_level"] is not None:
            return self.data["trans_level"]
        return None

    @property
    def strong_mode_enabled(self) -> bool:
        if self.firmware_version_minor == 25:
            if self.trans_level == 90:
                return True

        elif self.firmware_version_minor > 25 or self.firmware_version_minor == 0:
            return self.mode.value == "strong"

        return False

    @property
    def firmware_version(self) -> str:
        """Returns the fw_ver of miIO.info.

        For example 1.2.9_5033.
        """
        if self.device_info.firmware_version is None:
            raise AirHumidifierException("Missing firmware information")

        return self.device_info.firmware_version

    @property
    def firmware_version_major(self) -> str:
        parts = self.firmware_version.rsplit("_", 1)
        return parts[0]

    @property
    def firmware_version_minor(self) -> int:
        parts = self.firmware_version.rsplit("_", 1)
        try:
            return int(parts[1])
        except IndexError:
            return 0

    @property
    def motor_speed(self) -> Optional[int]:
        """Current fan speed."""
        if "speed" in self.data and self.data["speed"] is not None:
            return self.data["speed"]
        return None

    @property
    def depth(self) -> Optional[int]:
        """Return raw value of depth."""
        _LOGGER.warning(
            "The 'depth' property is deprecated and will be removed in the future. Use 'water_level' and 'water_tank_detached' properties instead."
        )
        if "depth" in self.data:
            return self.data["depth"]
        return None

    @property
    def water_level(self) -> Optional[int]:
        """Return current water level in percent.

        If water tank is full, depth is 125.
        """
        if self.depth is not None and self.depth <= 125:
            return int(self.depth / 1.25)
        return None

    @property
    def water_tank_detached(self) -> Optional[bool]:
        """True if the water tank is detached.

        If water tank is detached, depth is 127.
        """
        if self.data.get("depth") is not None:
            return self.data["depth"] == 127
        return None

    @property
    def dry(self) -> Optional[bool]:
        """Dry mode: The amount of water is not enough to continue to work for about 8
        hours.

        Return True if dry mode is on if available.
        """
        if "dry" in self.data and self.data["dry"] is not None:
            return self.data["dry"] == "on"
        return None

    @property
    def use_time(self) -> Optional[int]:
        """How long the device has been active in seconds."""
        return self.data["use_time"]

    @property
    def hardware_version(self) -> Optional[str]:
        """The hardware version."""
        return self.data["hw_version"]

    @property
    def button_pressed(self) -> Optional[str]:
        """Last pressed button."""
        if "button_pressed" in self.data and self.data["button_pressed"] is not None:
            return self.data["button_pressed"]
        return None


class AirHumidifier(Device):
    """Implementation of Xiaomi Mi Air Humidifier."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_HUMIDIFIER_V1,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_HUMIDIFIER_V1

        # TODO: convert to use generic device info in the future
        self.device_info: Optional[DeviceInfo] = None

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Temperature: {result.temperature} °C\n"
            "Humidity: {result.humidity} %\n"
            "LED brightness: {result.led_brightness}\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Target humidity: {result.target_humidity} %\n"
            "Trans level: {result.trans_level}\n"
            "Speed: {result.motor_speed}\n"
            "Depth: {result.depth}\n"
            "Water Level: {result.water_level} %\n"
            "Water tank detached: {result.water_tank_detached}\n"
            "Dry: {result.dry}\n"
            "Use time: {result.use_time}\n"
            "Hardware version: {result.hardware_version}\n"
            "Button pressed: {result.button_pressed}\n",
        )
    )
    def status(self) -> AirHumidifierStatus:
        """Retrieve properties."""
        if self.device_info is None:
            self.device_info = self.info()

        properties = AVAILABLE_PROPERTIES[self.model]

        # A single request is limited to 16 properties. Therefore the
        # properties are divided into multiple requests
        _props_per_request = 15

        # The CA1, CB1 and CB2 are limited to a single property per request
        if self.model in [
            MODEL_HUMIDIFIER_CA1,
            MODEL_HUMIDIFIER_CB1,
            MODEL_HUMIDIFIER_CB2,
        ]:
            _props_per_request = 1

        values = self.get_properties(properties, max_properties=_props_per_request)

        return AirHumidifierStatus(
            defaultdict(lambda: None, zip(properties, values)), self.device_info
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        try:
            return self.send("set_mode", [mode.value])
        except DeviceError as error:
            # {'code': -6011, 'message': 'device_poweroff'}
            if error.code == -6011:
                self.on()
                return self.send("set_mode", [mode.value])
            raise

    @command(
        click.argument("brightness", type=EnumType(LedBrightness)),
        default_output=format_output("Setting LED brightness to {brightness}"),
    )
    def set_led_brightness(self, brightness: LedBrightness):
        """Set led brightness."""
        if self.model in [MODEL_HUMIDIFIER_CA1, MODEL_HUMIDIFIER_CB1]:
            return self.send("set_led_b", [str(brightness.value)])

        return self.send("set_led_b", [brightness.value])

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        if led:
            return self.set_led_brightness(LedBrightness.Bright)
        else:
            return self.set_led_brightness(LedBrightness.Off)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        if buzzer:
            return self.send("set_buzzer", ["on"])
        else:
            return self.send("set_buzzer", ["off"])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        if lock:
            return self.send("set_child_lock", ["on"])
        else:
            return self.send("set_child_lock", ["off"])

    @command(
        click.argument("humidity", type=int),
        default_output=format_output("Setting target humidity to {humidity}"),
    )
    def set_target_humidity(self, humidity: int):
        """Set the target humidity."""
        if humidity not in [30, 40, 50, 60, 70, 80]:
            raise AirHumidifierException("Invalid target humidity: %s" % humidity)

        return self.send("set_limit_hum", [humidity])

    @command(
        click.argument("dry", type=bool),
        default_output=format_output(
            lambda dry: "Turning on dry mode" if dry else "Turning off dry mode"
        ),
    )
    def set_dry(self, dry: bool):
        """Set dry mode on/off."""
        if dry:
            return self.send("set_dry", ["on"])
        else:
            return self.send("set_dry", ["off"])


class AirHumidifierCA1(AirHumidifier):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_HUMIDIFIER_CA1
        )


class AirHumidifierCB1(AirHumidifier):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_HUMIDIFIER_CB1
        )


class AirHumidifierCB2(AirHumidifier):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_HUMIDIFIER_CB2
        )



from unittest import TestCase

import pytest

from miio import Heater
from miio.heater import MODEL_HEATER_ZA1, Brightness, HeaterException, HeaterStatus

from .dummies import DummyDevice


class DummyHeater(DummyDevice, Heater):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_HEATER_ZA1
        # This example response is just a guess. Please update!
        self.state = {
            "target_temperature": 24,
            "temperature": 22.1,
            "relative_humidity": 46,
            "poweroff_time": 0,
            "power": "on",
            "child_lock": "off",
            "buzzer": "on",
            "brightness": 1,
            "use_time": 0,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_target_temperature": lambda x: self._set_state(
                "target_temperature", x
            ),
            "set_brightness": lambda x: self._set_state("brightness", x),
            "set_buzzer": lambda x: self._set_state("buzzer", x),
            "set_child_lock": lambda x: self._set_state("child_lock", x),
            "set_poweroff_time": lambda x: self._set_state("poweroff_time", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def heater(request):
    request.cls.device = DummyHeater()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("heater")
class TestHeater(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(HeaterStatus(self.device.start_state))

        assert self.is_on() is True
        assert (
            self.state().target_temperature
            == self.device.start_state["target_temperature"]
        )
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().humidity == self.device.start_state["relative_humidity"]
        assert (
            self.state().delay_off_countdown == self.device.start_state["poweroff_time"]
        )
        assert self.state().child_lock is (
            self.device.start_state["child_lock"] == "on"
        )
        assert self.state().buzzer is (self.device.start_state["buzzer"] == "on")
        assert self.state().brightness == Brightness(
            self.device.start_state["brightness"]
        )
        assert self.state().use_time == self.device.start_state["use_time"]

    def test_set_target_temperature(self):
        def target_temperature():
            return self.device.status().target_temperature

        self.device.set_target_temperature(16)
        assert target_temperature() == 16
        self.device.set_target_temperature(24)
        assert target_temperature() == 24
        self.device.set_target_temperature(32)
        assert target_temperature() == 32

        with pytest.raises(HeaterException):
            self.device.set_target_temperature(15)

        with pytest.raises(HeaterException):
            self.device.set_target_temperature(33)

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(Brightness.Bright)
        assert brightness() == Brightness.Bright

        self.device.set_brightness(Brightness.Dim)
        assert brightness() == Brightness.Dim

        self.device.set_brightness(Brightness.Off)
        assert brightness() == Brightness.Off

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(0)
        assert delay_off_countdown() == 0
        self.device.delay_off(9)
        assert delay_off_countdown() == 9

        with pytest.raises(HeaterException):
            self.device.delay_off(-1)

        with pytest.raises(HeaterException):
            self.device.delay_off(9 * 3600 + 1)



import enum
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .exceptions import DeviceException
from .miot_device import DeviceStatus, MiotDevice, MiotMapping


class YeelightDualControlModuleException(DeviceException):
    pass


class Switch(enum.Enum):
    First = 0
    Second = 1


_MAPPING: MiotMapping = {
    # http://miot-spec.org/miot-spec-v2/instance?type=urn:miot-spec-v2:device:switch:0000A003:yeelink-sw1:1:0000C809
    # First Switch (siid=2)
    "switch_1_state": {"siid": 2, "piid": 1},  # bool
    "switch_1_default_state": {"siid": 2, "piid": 2},  # 0 - Off, 1 - On
    "switch_1_off_delay": {"siid": 2, "piid": 3},  # -1 - Off, [1, 43200] - delay in sec
    # Second Switch (siid=3)
    "switch_2_state": {"siid": 3, "piid": 1},  # bool
    "switch_2_default_state": {"siid": 3, "piid": 2},  # 0 - Off, 1 - On
    "switch_2_off_delay": {"siid": 3, "piid": 3},  # -1 - Off, [1, 43200] - delay in sec
    # Extensions (siid=4)
    "interlock": {"siid": 4, "piid": 1},  # bool
    "flex_mode": {"siid": 4, "piid": 2},  # 0 - Off, 1 - On
    "rc_list": {"siid": 4, "piid": 3},  # string
    "rc_list_for_del": {"siid": 4, "piid": 4},  # string
    "toggle": {"siid": 4, "piid": 5},  # 0 - First switch, 1 - Second switch
}


class DualControlModuleStatus(DeviceStatus):
    def __init__(self, data: Dict[str, Any]) -> None:
        """
        Response of Yeelight Dual Control Module
        {
            'id': 1,
            'result': [
                {'did': 'switch_1_state', 'siid': 2, 'piid': 1, 'code': 0, 'value': False},
                {'did': 'switch_1_default_state', 'siid': 2, 'piid': 2, 'code': 0, 'value': True},
                {'did': 'switch_1_off_delay', 'siid': 2, 'piid': 3, 'code': 0, 'value': 300},
                {'did': 'switch_2_state', 'siid': 3, 'piid': 1, 'code': 0, 'value': False},
                {'did': 'switch_2_default_state', 'siid': 3, 'piid': 2, 'code': 0, 'value': False},
                {'did': 'switch_2_off_delay', 'siid': 3, 'piid': 3, 'code': 0, 'value': 0},
                {'did': 'interlock', 'siid': 4, 'piid': 1, 'code': 0, 'value': False},
                {'did': 'flex_mode', 'siid': 4, 'piid': 2, 'code': 0, 'value': True},
                {'did': 'rc_list', 'siid': 4, 'piid': 2, 'code': 0, 'value': '[{"mac":"9db0eb4124f8","evtid":4097,"pid":339,"beaconkey":"3691bc0679eef9596bb63abf"}]'},
            ]
        }
        """
        self.data = data

    @property
    def switch_1_state(self) -> bool:
        """First switch state."""
        return bool(self.data["switch_1_state"])

    @property
    def switch_1_default_state(self) -> bool:
        """First switch default state."""
        return bool(self.data["switch_1_default_state"])

    @property
    def switch_1_off_delay(self) -> int:
        """First switch off delay."""
        return self.data["switch_1_off_delay"]

    @property
    def switch_2_state(self) -> bool:
        """Second switch state."""
        return bool(self.data["switch_2_state"])

    @property
    def switch_2_default_state(self) -> bool:
        """Second switch default state."""
        return bool(self.data["switch_2_default_state"])

    @property
    def switch_2_off_delay(self) -> int:
        """Second switch off delay."""
        return self.data["switch_2_off_delay"]

    @property
    def interlock(self) -> bool:
        """Interlock."""
        return bool(self.data["interlock"])

    @property
    def flex_mode(self) -> int:
        """Flex mode."""
        return self.data["flex_mode"]

    @property
    def rc_list(self) -> str:
        """List of paired remote controls."""
        return self.data["rc_list"]


class YeelightDualControlModule(MiotDevice):
    """Main class representing the Yeelight Dual Control Module (yeelink.switch.sw1)
    which uses MIoT protocol."""

    mapping = _MAPPING

    @command(
        default_output=format_output(
            "",
            "First Switch Status: {result.switch_1_state}\n"
            "First Switch Default State: {result.switch_1_default_state}\n"
            "First Switch Delay: {result.switch_1_off_delay}\n"
            "Second Switch Status: {result.switch_2_state}\n"
            "Second Switch Default State: {result.switch_2_default_state}\n"
            "Second Switch Delay: {result.switch_2_off_delay}\n"
            "Interlock: {result.interlock}\n"
            "Flex Mode: {result.flex_mode}\n"
            "RC list: {result.rc_list}\n",
        )
    )
    def status(self) -> DualControlModuleStatus:
        """Retrieve properties."""
        p = [
            "switch_1_state",
            "switch_1_default_state",
            "switch_1_off_delay",
            "switch_2_state",
            "switch_2_default_state",
            "switch_2_off_delay",
            "interlock",
            "flex_mode",
            "rc_list",
        ]
        """Filter only readable properties for status"""
        properties = [
            {"did": k, **v}
            for k, v in filter(lambda item: item[0] in p, _MAPPING.items())
        ]
        values = self.get_properties(properties)
        return DualControlModuleStatus(
            dict(map(lambda v: (v["did"], v["value"]), values))
        )

    @command(
        click.argument("switch", type=EnumType(Switch)),
        default_output=format_output("Turn {switch} switch on"),
    )
    def on(self, switch: Switch):
        """Turn switch on."""
        if switch == Switch.First:
            return self.set_property("switch_1_state", True)
        elif switch == Switch.Second:
            return self.set_property("switch_2_state", True)

    @command(
        click.argument("switch", type=EnumType(Switch)),
        default_output=format_output("Turn {switch} switch off"),
    )
    def off(self, switch: Switch):
        """Turn switch off."""
        if switch == Switch.First:
            return self.set_property("switch_1_state", False)
        elif switch == Switch.Second:
            return self.set_property("switch_2_state", False)

    @command(
        click.argument("switch", type=EnumType(Switch)),
        default_output=format_output("Toggle {switch} switch"),
    )
    def toggle(self, switch: Switch):
        """Toggle switch."""
        return self.set_property("toggle", switch.value)

    @command(
        click.argument("state", type=bool),
        click.argument("switch", type=EnumType(Switch)),
        default_output=format_output("Set {switch} switch default state to: {state}"),
    )
    def set_default_state(self, state: bool, switch: Switch):
        """Set switch default state."""
        if switch == Switch.First:
            return self.set_property("switch_1_default_state", int(state))
        elif switch == Switch.Second:
            return self.set_property("switch_2_default_state", int(state))

    @command(
        click.argument("delay", type=int),
        click.argument("switch", type=EnumType(Switch)),
        default_output=format_output("Set {switch} switch off delay to {delay} sec."),
    )
    def set_switch_off_delay(self, delay: int, switch: Switch):
        """Set switch off delay, should be between -1 to 43200 (in seconds)"""
        if delay < -1 or delay > 43200:
            raise YeelightDualControlModuleException(
                "Invalid switch delay: %s (should be between -1 to 43200)" % delay
            )

        if switch == Switch.First:
            return self.set_property("switch_1_off_delay", delay)
        elif switch == Switch.Second:
            return self.set_property("switch_2_off_delay", delay)

    @command(
        click.argument("flex_mode", type=bool),
        default_output=format_output("Set flex mode to: {flex_mode}"),
    )
    def set_flex_mode(self, flex_mode: bool):
        """Set flex mode."""
        return self.set_property("flex_mode", int(flex_mode))

    @command(
        click.argument("rc_mac", type=str),
        default_output=format_output("Delete remote control with MAC: {rc_mac}"),
    )
    def delete_rc(self, rc_mac: str):
        """Delete remote control by MAC."""
        return self.set_property("rc_list_for_del", rc_mac)

    @command(
        click.argument("interlock", type=bool),
        default_output=format_output("Set interlock to: {interlock}"),
    )
    def set_interlock(self, interlock: bool):
        """Set interlock."""
        return self.set_property("interlock", interlock)
